''' Transistor elements '''

from __future__ import annotations
from typing import Optional

__all__ = [
    'Bjt', 'Bjt2', 'BjtNpn', 'BjtNpn2', 'BjtPnp', 'BjtPnp2', 'BjtPnp2c', 'BjtPnp2c2',
    'JFet', 'JFet2', 'JFetN', 'JFetN2', 'JFetP', 'JFetP2',
    'NFet', 'NFet2', 'NMos', 'NMos2', 'PFet', 'PFet2', 'PMos', 'PMos2',
    "AnalogNFet", "AnalogPFet", "AnalogBiasedFet"]

from .elements import Element, Element2Term, gap
from .twoterm import reswidth
from ..segments import Segment, SegmentPoly, SegmentCircle
from ..types import Point


class _Mosfet(Element):
    '''Base Class for Metal Oxide Semiconductor Field Effect Transistors

        Args:
            variant: one of {'nmos', 'pmos'}
            diode: Draw body diode [default: False]
            circle: Draw circle around the mosfet [default: False]

        Anchors:
            * source
            * drain
            * gate

    Note: vertical orientation.  For horizontal orientation, see _Mosfet2.
    '''
    _element_defaults = {
        'diode': False,
        'circle': False
    }
    __variants = ['nmos', 'pmos']

    def __init__(self,
                 variant: str,
                 *,
                 diode: Optional[bool] = False,
                 circle: Optional[bool] = False,
                 **kwargs):
        if variant not in self.__variants:
            raise ValueError(
                "Parameter 'variant' must be one of {}, not {}.".format(
                    self.__variants, variant))

        super().__init__(**kwargs)
        self.elmparams['ilabel'] = 'right'  # Draw current labels on this side
        u = reswidth*0.5

        # gate
        self.segments.extend([
            Segment([(-10*u, -6*u), (-5*u, -6*u)]),
            Segment([(-5*u, -6*u), (-5*u, -14*u)]),
            ])

        # ---
        self.segments.extend([
            Segment([(-3*u, -6.5*u), (-3*u, -7.5*u)]),
            Segment([(-3*u, -9.5*u), (-3*u, -10.5*u)]),
            Segment([(-3*u, -12.5*u), (-3*u, -13.5*u)]),
            ])

        # top lead
        self.segments.extend([
            Segment([(0, -7*u), (0, 0)]),
            Segment([(-3*u, -7*u), (0, -7*u)]),
            ])

        # bottom lead
        self.segments.extend([
            Segment([(0, -20*u), (0, -13*u)]),
            Segment([(-3*u, -13*u), (0, -13*u)]),
            ])

        if variant == 'nmos':

            # source
            self.segments.extend([
                Segment(
                    [(-3*u, -10*u), (0, -10*u)],
                    arrow='<-', arrowwidth=2*u, arrowlength=2*u),
                Segment([(0, -10*u), (0, -13*u)]),
                ])

            self.anchors['drain'] = (0, 0)
            self.anchors['gate'] = (-10*u, -6*u)
            self.anchors['center'] = (0, -10*u)
            self.anchors['source'] = (0, -20*u)

        elif variant == 'pmos':

            # source
            self.segments.extend([
                Segment(
                    [(-3*u, -10*u), (0, -10*u)],
                    arrow='->', arrowwidth=2*u, arrowlength=2*u),
                Segment([(0, -10*u), (0, -7*u)]),
                ])

            self.anchors['source'] = (0, 0)
            self.anchors['gate'] = (-10*u, -6*u)
            self.anchors['center'] = (0, -10*u)
            self.anchors['drain'] = (0, -20*u)

        self.elmparams['drop'] = (0, -20*u)
        self.elmparams['lblloc'] = 'rgt'

        if self.params['diode']:
            self.segments.extend([
                Segment([(0, -7*u), (3*u, -7*u)]),
                Segment([(3*u, -7*u), (3*u, -9*u)]),
                Segment([(2*u, -9*u), (4*u, -9*u)]),
                SegmentPoly([(3*u, -9*u), (2*u, -11*u), (4*u, -11*u)]),
                Segment([(3*u, -11*u), (3*u, -13*u)]),
                Segment([(3*u, -13*u), (0, -13*u)]),
                ])

        if self.params['circle']:
            self.segments.append(
                SegmentCircle((-1*u, -10*u), 7*u))


class NMos(_Mosfet):
    ''' N-type Metal Oxide Semiconductor Field Effect Transistor

        Args:
            diode: Draw body diode
            circle: Draw circle around the mosfet

        Anchors:
            * source
            * drain
            * gate

    Note: vertical orientation.  For horizontal orientation, see NMos2.
    '''
    def __init__(self,
                 diode: Optional[bool] = False,
                 circle: Optional[bool] = False,
                 **kwargs):
        super().__init__(variant='nmos', diode=diode, circle=circle, **kwargs)


class PMos(_Mosfet):
    ''' P-type Metal Oxide Semiconductor Field Effect Transistor

        Args:
            diode: Draw body diode
            circle: Draw circle around the mosfet

        Anchors:
            * source
            * drain
            * gate

    Note: vertical orientation.  For horizontal orientation, see PMos2.
    '''

    def __init__(self,
                 diode: Optional[bool] = False,
                 circle: Optional[bool] = False,
                 **kwargs):
        super().__init__(variant='pmos', diode=diode, circle=circle, **kwargs)


class _Mosfet2(Element2Term):
    '''Base Class for Metal Oxide Semiconductor Field Effect Transistors

        Args:
            variant: one of {'nmos', 'pmos'}
            diode: Draw body diode
            circle: Draw circle around the mosfet

        Anchors:
            * source
            * drain
            * gate

    Note: horizontal orientation.  For vertical orientation, see _Mosfet.
    '''
    _element_defaults = {
        'diode': False,
        'circle': False
    }
    __variants = ['nmos', 'pmos']

    def __init__(self, variant: str, *,
                 diode: Optional[bool] = False,
                 circle: Optional[bool] = False,
                 **kwargs):
        if variant not in self.__variants:
            raise ValueError(
                "Parameter 'variant' must be one of {}, not {}.".format(
                    self.__variants, variant))

        super().__init__(**kwargs)
        self.variant = variant
        u = reswidth*0.5

        # top and bottom leads
        self.segments.append(
            Segment([
                (0, 0), (7*u, 0), (7*u, -3*u), gap,
                (13*u, -3*u), (13*u, 0), (20*u, 0),
                ]))

        # gate
        self.segments.extend([
            Segment([(6*u, -10*u), (6*u, -5*u)]),
            Segment([(6*u, -5*u), (14*u, -5*u)]),
            ])

        # ---
        self.segments.extend([
            Segment([(6.5*u, -3*u), (7.5*u, -3*u)]),
            Segment([(9.5*u, -3*u), (10.5*u, -3*u)]),
            Segment([(12.5*u, -3*u), (13.5*u, -3*u)]),
            ])

        if variant == 'nmos':

            # source
            self.segments.extend([
                Segment(
                    [(10*u, -3*u), (10*u, 0)],
                    arrow='<-', arrowwidth=2*u, arrowlength=2*u),
                Segment([(10*u, 0), (13*u, 0)]),
                ])

            self.anchors['drain'] = (0, 0)
            self.anchors['gate'] = (6*u, -10*u)
            self.anchors['source'] = (20*u, 0)

        elif variant == 'pmos':

            # source
            self.segments.extend([
                Segment(
                    [(10*u, -3*u), (10*u, 0)],
                    arrow='->', arrowwidth=2*u, arrowlength=2*u),
                Segment([(10*u, 0), (7*u, 0)]),
                ])

            self.anchors['source'] = (0, 0)
            self.anchors['gate'] = (6*u, -10*u)
            self.anchors['drain'] = (20*u, 0)

        self.anchors['center'] = (10*u, 0)
        self.elmparams['drop'] = (20*u, 0)
        self.elmparams['lblloc'] = 'bottom'

        if self.params['diode']:
            self.segments.extend([
                Segment([(7*u, 0), (7*u, 3*u)]),
                Segment([(7*u, 3*u), (9*u, 3*u)]),
                Segment([(9*u, 2*u), (9*u, 4*u)]),
                SegmentPoly([(9*u, 3*u), (11*u, 2*u), (11*u, 4*u)]),
                Segment([(11*u, 3*u), (13*u, 3*u)]),
                Segment([(13*u, 3*u), (13*u, 0)]),
                ])

        if self.params['circle']:
            self.segments.append(
                SegmentCircle((10*u, -1*u), 7*u))


class NMos2(_Mosfet2):
    ''' N-type Metal Oxide Semiconductor Field Effect Transistor

        Args:
            diode: Draw body diode
            circle: Draw circle around the mosfet

        Anchors:
            * source
            * drain
            * gate

    Note: horizontal orientation.  For vertical orientation, see NMos.
    '''
    def __init__(self,
                 diode: Optional[bool] = False,
                 circle: Optional[bool] = False,
                 **kwargs):
        super().__init__(variant='nmos', diode=diode, circle=circle, **kwargs)


class PMos2(_Mosfet2):
    ''' P-type Metal Oxide Semiconductor Field Effect Transistor

        Args:
            diode: Draw body diode
            circle: Draw circle around the mosfet

        Anchors:
            * source
            * drain
            * gate

    Note: horizontal orientation.  For vertical orientation, see PMos.
    '''
    def __init__(self,
                 diode: Optional[bool] = False,
                 circle: Optional[bool] = False,
                 **kwargs):
        super().__init__(variant='pmos', diode=diode, circle=circle, **kwargs)


fetw = reswidth * 4
feth = reswidth * 5
fetl = feth / 2
fetgap = reswidth
fetr = reswidth * .7  # Radius of "not" bubble


class NFet(Element):
    ''' N-type Field Effect Transistor

        Args:
            bulk: Draw bulk contact [default: False]

        Anchors:
            * source
            * drain
            * gate
            * center
    '''
    _element_defaults = {
        'bulk': False
    }
    def __init__(self, *, bulk: Optional[bool] = None, **kwargs):
        super().__init__(**kwargs)
        self.elmparams['ilabel'] = 'left'  # Draw current labels on this side

        self.segments.append(Segment([(0, 0), (0, -fetl), (fetw, -fetl),
                                      (fetw, -fetl-fetw), (0, -fetl-fetw),
                                      (0, -2*fetl-fetw)]))
        self.segments.append(Segment([(fetw+fetgap, -fetl),
                                      (fetw+fetgap, -fetl-fetw)]))
        self.segments.append(Segment([(fetw+fetgap, -fetl-fetw/2),
                                      (fetw+fetgap+fetl+fetr, -fetl-fetw/2)]))
        self.anchors['source'] = (0, -2*fetl-fetw)
        self.anchors['drain'] = (0, 0)
        self.anchors['gate'] = (fetw+fetgap+fetl+fetr, -fetl-fetw/2)
        self.anchors['center'] = (0, -fetl - fetw/2)
        self.elmparams['drop'] = (0, -2*fetl-fetw)
        self.elmparams['lblloc'] = 'lft'
        if self.params['bulk']:
            self.segments.append(Segment([(0, -fetl-fetw/2), (fetw, -fetl-fetw/2)],
                                         arrow='->', arrowwidth=.2))
            self.anchors['bulk'] = (0, -fetl-fetw/2)


class PFet(Element):
    ''' P-type Field Effect Transistor

        Args:
            bulk: Draw bulk contact

        Anchors:
            source
            drain
            gate
            center
    '''
    _element_defaults = {
        'bulk': False
    }
    def __init__(self, *, bulk: Optional[bool] = None, **kwargs):
        super().__init__(**kwargs)
        self.elmparams['ilabel'] = 'left'  # Draw current labels on this side

        self.segments.append(Segment([(0, 0), (0, -fetl), (fetw, -fetl),
                                      (fetw, -fetl-fetw), (0, -fetl-fetw),
                                      (0, -2*fetl-fetw)]))
        self.segments.append(Segment([(fetw+fetgap, -fetl),
                                      (fetw+fetgap, -fetl-fetw)]))
        self.segments.append(Segment([(fetw+fetgap+fetr*2, -fetl-fetw/2),
                                      (fetw+fetgap+fetl+fetr, -fetl-fetw/2)]))
        self.segments.append(SegmentCircle((fetw+fetgap+fetr, -fetl-fetw/2), fetr))

        self.anchors['source'] = (0, 0)
        self.anchors['drain'] = (0, -2*fetl-fetw)
        self.anchors['gate'] = (fetw+fetgap+fetl+fetr, -fetl-fetw/2)
        self.anchors['center'] = (0, -fetl - fetw / 2)
        self.elmparams['drop'] = (0, -2*fetl-fetw)
        self.elmparams['lblloc'] = 'lft'
        if self.params['bulk']:
            self.segments.append(Segment([(0, -fetl-fetw/2), (fetw, -fetl-fetw/2)],
                                         arrow='->', arrowwidth=.2))
            self.anchors['bulk'] = (0, -fetl-fetw/2)


class NFet2(Element2Term):
    ''' N-type Field Effect Transistor which extends
        source/drain leads to the desired length

        Args:
            bulk: Draw bulk contact [default: False]

        Anchors:
            * source
            * drain
            * gate
    '''
    _element_defaults = {
        'bulk': False
        }
    def __init__(self, *, bulk: Optional[bool] = None, **kwargs):
        super().__init__(**kwargs)
        self.elmparams['ilabel'] = 'bottom'  # Draw current labels on this side

        self.segments.append(Segment([(0, 0), (fetl, 0), (fetl, fetw),
                                      (fetl+fetw, fetw), (fetl+fetw, 0), (2*fetl+fetw, 0)]))
        self.segments.append(Segment([(fetl, fetw+fetgap), (fetl+fetw, fetw+fetgap)]))
        self.segments.append(Segment([(fetl+fetw/2, fetw+fetgap), (fetl+fetw/2, fetw+fetgap+fetr+fetl)]))
        self.anchors['isource'] = (2*fetl+fetw, 0)
        self.anchors['idrain'] = (0, 0)
        self.anchors['gate'] = (fetl+fetw/2, fetw+fetgap+fetr+fetl)
        self.elmparams['lblloc'] = 'bottom'
        if self.params['bulk']:
            self.segments.append(Segment([(fetl+fetw/2, 0), (fetl+fetw/2, fetw)],
                                         arrow='->', arrowwidth=.2))
            self.anchors['bulk'] = (fetl+fetw/2, 0)

    def _place_anchors(self, start, end):
        ''' Allow positioning anchors relative to extended endpoints
            before the element is placed
        '''
        super()._place_anchors(start, end)
        self.anchors['source'] = self.anchors['end']
        self.anchors['drain'] = self.anchors['start']
        if self._userparams.get('reverse', False):
            self.anchors['source'] = self.anchors['start']
            self.anchors['drain'] = self.anchors['end']


class PFet2(Element2Term):
    ''' P-type Field Effect Transistor which extends
        source/drain leads to the desired length

        Args:
            bulk: Draw bulk contact [default: False]

        Anchors:
            * source
            * drain
            * gate
    '''
    _element_defaults = {
        'bulk': False
    }
    def __init__(self, *, bulk: Optional[bool] = None, **kwargs):
        super().__init__(**kwargs)
        self.elmparams['ilabel'] = 'bottom'  # Draw current labels on this side

        self.segments.append(Segment([(0, 0), (fetl, 0), (fetl, fetw),
                                      (fetl+fetw, fetw), (fetl+fetw, 0), (2*fetl+fetw, 0)]))
        self.segments.append(Segment([(fetl, fetw+fetgap), (fetl+fetw, fetw+fetgap)]))
        self.segments.append(SegmentCircle((fetl+fetw/2, fetw+fetgap+fetr), fetr))
        self.segments.append(Segment([(fetl+fetw/2, fetw+fetgap+fetr*2), (fetl+fetw/2, fetw+fetgap+fetr+fetl)]))
        self.anchors['isource'] = (2*fetl+fetw, 0)
        self.anchors['idrain'] = (0, 0)
        self.anchors['gate'] = (fetl+fetw/2, fetw+fetgap+fetr+fetl)
        self.elmparams['lblloc'] = 'bottom'
        if self.params['bulk']:
            self.segments.append(Segment([(fetl+fetw/2, 0), (fetl+fetw/2, fetw)],
                                         arrow='->', arrowwidth=.2))
            self.anchors['bulk'] = (fetl+fetw/2, 0)

    def _place_anchors(self, start, end):
        ''' Allow positioning anchors relative to extended endpoints
            before the element is placed
        '''
        super()._place_anchors(start, end)
        self.anchors['source'] = self.anchors['end']
        self.anchors['drain'] = self.anchors['start']
        if self._userparams.get('reverse', False):
            self.anchors['source'] = self.anchors['start']
            self.anchors['drain'] = self.anchors['end']


# Analog style FETs
afetw = reswidth * 2.5
afeth = afetw * 2
afetl = afeth * 0.5
afetgap = afetw * 0.2
afeti = afeth * 0.1   # gate inset
afetb = afeti * 0.75  # bias dot radius
afeta = 0.25          # fet arrow head width


class _AnalogFet(Element):
    _element_defaults = {
        'bulk': False,
        'offset_gate': True,
        'arrow': True
    }
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.elmparams['ilabel'] = 'left'  # Draw current labels on this side


class AnalogNFet(_AnalogFet):
    ''' N-type Field Effect Transistor, analog style

        Args:
            bulk: Draw bulk contact
            offset_gate: Draw gate on the source side of the transistor, rather than middle
            arrow: Draw source arrow on the transistor if bulk arrow is not drawn

        Anchors:
            source
            drain
            gate
            bulk (if bulk=True)
            center
    '''
    def __init__(self, *,
                 bulk: Optional[bool] = None,
                 offset_gate: Optional[bool] = None,
                 arrow: Optional[bool] = None,
                 **kwargs):
        super().__init__(**kwargs)
        _arrow = self.params['arrow'] if not self.params['bulk'] else False
        self.segments.append(Segment([(0, 0), (0, -afetl), (afetw, -afetl),
                                      (afetw, -afetl - afeth), (0, -afetl - afeth),
                                      (0, -2 * afetl - afeth)]))
        if _arrow:
            self.segments.append(Segment([(afetw, -afetl - afeth), (0, -afetl - afeth)],
                                         arrow='->', arrowwidth=afeta, arrowlength=afeta))
        self.segments.append(Segment([(afetw + afetgap, -afetl - afeti),
                                      (afetw + afetgap, -afetl - afeth + afeti)]))

        if self.params['offset_gate']:
            self.segments.append(Segment([(afetw + afetgap, -afetl - afeth + afeti),
                                          (afetw + afetgap + afetl, -afetl - afeth + afeti)]))
            self.anchors['gate'] = (afetw + afetgap + afetl, -afetl - afeth + afeti)
        else:
            self.segments.append(Segment([(afetw + afetgap, -afetl - afeth / 2),
                                          (afetw + afetgap + afetl, -afetl - afeth / 2)]))
            self.anchors['gate'] = (afetw + afetgap + afetl, -afetl - afeth / 2)

        self.anchors['source'] = (0, -2 * afetl - afeth)
        self.anchors['drain'] = (0, 0)
        self.anchors['center'] = (0, -afetl - afeth / 2)
        self.elmparams['drop'] = (0, -2 * afetl - afeth)
        self.elmparams['lblloc'] = 'lft'

        if self.params['bulk']:
            self.segments.append(Segment([(0, -afetl - afeth / 2), (afetw, -afetl - afeth / 2)],
                                         arrow='->', arrowwidth=afeta, arrowlength=afeta))
            self.anchors['bulk'] = (0, -afetl - afeth / 2)


class AnalogPFet(_AnalogFet):
    ''' P-type Field Effect Transistor, analog style

        Args:
            bulk: Draw bulk contact
            offset_gate: Draw gate on the source side of the transistor, rather than middle
            arrow: Draw source arrow on the transistor if bulk arrow is not drawn

        Anchors:
            source
            drain
            gate
            bulk (if bulk=True)
            center
    '''
    def __init__(self, *,
                 bulk: Optional[bool] = None,
                 offset_gate: Optional[bool] = None,
                 arrow: Optional[bool] = None,
                 **kwargs):
        super().__init__(**kwargs)
        _arrow = self.params['arrow'] if not self.params['bulk'] else False
        self.segments.append(Segment([(0, 0), (0, -afetl), (afetw, -afetl),
                                      (afetw, -afetl - afeth), (0, -afetl - afeth),
                                      (0, -2 * afetl - afeth)]))
        if _arrow:
            self.segments.append(Segment([(0, -afetl), (afetw, -afetl)],
                                         arrow='->', arrowwidth=afeta, arrowlength=afeta))
        self.segments.append(Segment([(afetw + afetgap, -afetl - afeti),
                                      (afetw + afetgap, -afetl - afeth + afeti)]))
        if self.params['offset_gate']:
            self.segments.append(Segment([(afetw + afetgap, -afetl - afeti),
                                          (afetw + afetgap + afetl, -afetl - afeti)]))
            self.anchors['gate'] = (afetw + afetgap + afetl, -afetl - afeti)
        else:
            self.segments.append(Segment([(afetw + afetgap, -afetl - afeth / 2),
                                          (afetw + afetgap + afetl, -afetl - afeth / 2)]))
            self.anchors['gate'] = (afetw + afetgap + afetl, -afetl - afeth / 2)

        self.anchors['source'] = (0, 0)
        self.anchors['drain'] = (0, -2 * afetl - afeth)
        self.anchors['center'] = (0, -afetl - afeth / 2)
        self.elmparams['drop'] = (0, -2 * afetl - afeth)
        self.elmparams['lblloc'] = 'lft'
        if self.params['bulk']:
            self.segments.append(Segment([(afetw, -afetl - afeth / 2), (0, -afetl - afeth / 2)],
                                         arrow='->', arrowwidth=afeta, arrowlength=afeta))
            self.anchors['bulk'] = (0, -afetl - afeth / 2)


class AnalogBiasedFet(_AnalogFet):
    ''' Generic biased small-signal Field Effect Transistor, analog style

        Args:
            bulk: Draw bulk contact
            offset_gate: Draw gate on the source side of the transistor, rather than middle
            arrow: Draw source dot on the transistor if bulk dot is not drawn

        Anchors:
            source
            drain
            gate
            bulk (if bulk=True)
            center
    '''
    def __init__(self, *,
                 bulk: Optional[bool] = None,
                 offset_gate: Optional[bool] = None,
                 arrow: Optional[bool] = None,
                 **kwargs):
        super().__init__(**kwargs)
        _arrow = self.params['arrow'] if not self.params['bulk'] else False
        self.segments.append(Segment([(0, 0), (0, -afetl), (afetw, -afetl),
                                      (afetw, -afetl - afeth), (0, -afetl - afeth),
                                      (0, -2 * afetl - afeth)]))
        if _arrow:
            self.segments.append(SegmentCircle(center=(afetb * 2, -afetl - afeth), radius=afetb, fill=True))
        self.segments.append(Segment([(afetw + afetgap, -afetl - afeti),
                                      (afetw + afetgap, -afetl - afeth + afeti)]))

        if self.params['offset_gate']:
            self.segments.append(Segment([(afetw + afetgap, -afetl - afeth + afeti),
                                          (afetw + afetgap + afetl, -afetl - afeth + afeti)]))
            self.anchors['gate'] = (afetw + afetgap + afetl, -afetl - afeth + afeti)
        else:
            self.segments.append(Segment([(afetw + afetgap, -afetl - afeth / 2),
                                          (afetw + afetgap + afetl, -afetl - afeth / 2)]))
            self.anchors['gate'] = (afetw + afetgap + afetl, -afetl - afeth / 2)

        self.anchors['source'] = (0, -2 * afetl - afeth)
        self.anchors['drain'] = (0, 0)
        self.anchors['center'] = (0, -afetl - afeth / 2)
        self.elmparams['drop'] = (0, -2 * afetl - afeth)
        self.elmparams['lblloc'] = 'lft'
        if self.params['bulk']:
            self.segments.append(Segment([(0, -afetl - afeth / 2),
                                          (afetw, -afetl - afeth / 2)]))
            self.segments.append(SegmentCircle(center=(afetw - afetb * 2, -afetl - afeth / 2), radius=afetb, fill=True))
            self.anchors['bulk'] = (0, -afetl - afeth / 2)


# Junction FETs
fete = fetw*.2  # JFET extension
jfetw = reswidth*3


class JFet(Element):
    ''' Junction Field Effect Transistor (untyped)

        Anchors:
            * source
            * drain
            * gate
            * center
    '''
    _element_defaults = {
        'circle': False,
        'lblloc': 'lft'
    }
    def __init__(self, *, circle: Optional[bool] = None, **kwargs):
        super().__init__(**kwargs)
        self.elmparams['ilabel'] = 'left'  # Draw current labels on this side
        self.segments.append(Segment(
            [(0, 0), (0, -fetl), (jfetw, -fetl), (jfetw, -fetl+fete),
             (jfetw, -fetl-jfetw-fete), (jfetw, -fetl-jfetw),
             (0, -fetl-jfetw), (0, -2*fetl-jfetw)]))
        self.segments.append(Segment([(jfetw, -fetl-jfetw), (jfetw+fetl, -fetl-jfetw)]))
        self.elmparams['drop'] = (jfetw+fetl, -fetl-jfetw)
        self.anchors['source'] = (0, -2*fetl-jfetw)
        self.anchors['drain'] = (0, 0)
        self.anchors['gate'] = (jfetw+fetl, -fetl-jfetw)
        self.anchors['center'] = (0, -fetl-jfetw/2)
        if self.params['circle']:
            self.segments.append(SegmentCircle((jfetw/2, -fetw), fetw*1.1))


class JFetN(JFet):
    ''' N-type Junction Field Effect Transistor

        Args:
            circle: Draw circle around the transistor

        Anchors:
            * source
            * drain
            * gate
            * center
    '''
    def __init__(self, *, circle: Optional[bool] = None, **kwargs):
        super().__init__(circle=circle, **kwargs)
        self.segments.append(Segment([(jfetw+.1, -fetl-jfetw), (jfetw+.3, -fetl-jfetw)],
                                     arrow='->', arrowwidth=.2, arrowlength=.2))


class JFet2(Element2Term):
    ''' Junction Field Effect Transistor (untyped) which extends
        collector/emitter leads to the desired length

        Anchors:
            * source
            * drain
            * gate
    '''
    _element_defaults = {
        'circle': False,
        'lblloc': 'bottom'
    }
    def __init__(self, *, circle: Optional[bool] = None, **kwargs):
        super().__init__(**kwargs)
        self.elmparams['ilabel'] = 'bottom'  # Draw current labels on this side

        self.segments.append(Segment([
            (0, 0), (fetl, 0), (fetl, jfetw), (fetl+jfetw, jfetw),
            (fetl+jfetw, 0), (2*fetl+jfetw, 0)]))
        self.segments.append(Segment([(fetl-fete, jfetw), (fetl+jfetw+fete, jfetw)]))
        self.segments.append(Segment([(fetl+jfetw, jfetw), (fetl+jfetw, jfetw+fetl)]))
        self.anchors['isource'] = (2*fetl+jfetw, 0)
        self.anchors['idrain'] = (0, 0)
        self.anchors['gate'] = (fetl+jfetw, jfetw+fetl)

        if self.params['circle']:
            self.segments.append(SegmentCircle((fetl+jfetw/2, jfetw/2), fetw*1.1))

    def _place_anchors(self, start, end):
        ''' Allow positioning anchors relative to extended endpoints
            before the element is placed
        '''
        super()._place_anchors(start, end)
        self.anchors['source'] = self.anchors['end']
        self.anchors['drain'] = self.anchors['start']
        if self._userparams.get('reverse', False):
            self.anchors['source'] = self.anchors['start']
            self.anchors['drain'] = self.anchors['end']


class JFetN2(JFet2):
    ''' N-type Junction Field Effect Transistor which extends
        collector/emitter leads to the desired length

        Args:
            circle: Draw circle around the transistor

        Anchors:
            * source
            * drain
            * gate
    '''
    def __init__(self, *, circle: Optional[bool] = None, **kwargs):
        super().__init__(circle=circle, **kwargs)
        self.segments.append(Segment([(fetl+jfetw, jfetw), (fetl+jfetw, jfetw+0.3)],
                                     arrow='->', arrowwidth=.2, arrowlength=.2))


class JFetP2(JFet2):
    ''' P-type Junction Field Effect Transistor which extends
        collector/emitter leads to the desired length

        Args:
            circle: Draw circle around the transistor

        Anchors:
            * source
            * drain
            * gate
    '''
    def __init__(self, *, circle: Optional[bool] = None, **kwargs):
        super().__init__(circle=circle, **kwargs)
        self.segments.append(Segment([(fetl+jfetw, jfetw+0.3), (fetl+jfetw, jfetw)],
                                     arrow='->', arrowwidth=.2, arrowlength=.2))


class JFetP(JFet):
    ''' P-type Junction Field Effect Transistor

        Args:
            circle: Draw circle around the transistor

        Anchors:
            * source
            * drain
            * gate
            * center
    '''
    def __init__(self, *, circle: Optional[bool] = None, **kwargs):
        super().__init__(circle=circle, **kwargs)
        self.segments.append(Segment([(jfetw+.25, -fetl-jfetw), (jfetw, -fetl-jfetw)],
                                     arrow='->', arrowwidth=.2, arrowlength=.2))


# BJT transistors
bjt_r = reswidth*3.3   # Radius of BJT circle
bjt_v = bjt_r*2/3  # x coord of vertical line
bjt_v_len = bjt_r*4/3  # height of vertical line
bjt_a = bjt_v_len/4    # Intercept of emitter/collector lines
bjt_emx = bjt_v + bjt_r*.7  # x-coord of emitter exiting circle
bjt_emy = bjt_v_len*.7    # y-coord of emitter exiting circle


class Bjt(Element):
    ''' Bipolar Junction Transistor (untyped)

        Args:
            circle: Draw circle around the transistor

        Anchors:
            * collector
            * emitter
            * base
            * center
    '''
    _element_defaults = {
        'circle': False,
        'lblloc': 'rgt'
    }
    def __init__(self, *, circle: Optional[bool] = None, **kwargs):
        super().__init__(**kwargs)
        self.elmparams['ilabel'] = 'right'  # Draw current labels on this side

        self.segments.append(Segment([(0, 0), (bjt_v, 0)]))
        self.segments.append(Segment([(bjt_v, bjt_v_len/2), (bjt_v, -bjt_v_len/2)]))
        self.segments.append(Segment([(bjt_v, bjt_a), (bjt_emx, bjt_emy),
                                      (bjt_emx, bjt_emy+bjt_a)]))
        self.segments.append(Segment([(bjt_v, -bjt_a), (bjt_emx, -bjt_emy),
                                      (bjt_emx, -bjt_emy-bjt_a)]))
        if self.params['circle']:
            self.segments.append(SegmentCircle((bjt_r, 0), bjt_r))
        self.elmparams['drop'] = (bjt_emx, bjt_emy+bjt_a)
        self.anchors['base'] = (0, 0)
        self.anchors['collector'] = (bjt_emx, bjt_emy+bjt_a)
        self.anchors['emitter'] = (bjt_emx, -bjt_emy-bjt_a)
        self.anchors['center'] = (bjt_emx, 0)
        self.base: Point
        self.collector: Point
        self.emitter: Point


class BjtNpn(Bjt):
    ''' NPN Bipolar Junction Transistor

        Args:
            circle: Draw circle around the transistor

        Anchors:
            * collector
            * emitter
            * base
            * center
    '''
    def __init__(self, *, circle: Optional[bool] = None, **kwargs):
        super().__init__(circle=circle, **kwargs)
        self.segments.append(Segment([(bjt_v, -bjt_a), (bjt_emx, -bjt_emy)],
                                     arrow='->', arrowwidth=.2))


class BjtPnp(Bjt):
    ''' PNP Bipolar Junction Transistor

        Args:
            circle: Draw circle around the transistor

        Anchors:
            * collector
            * emitter
            * base
            * center
    '''
    def __init__(self, *, circle: Optional[bool] = None, **kwargs):
        super().__init__(circle=circle, **kwargs)
        self.segments.append(Segment([(bjt_emx, bjt_emy), (bjt_v, bjt_a)], arrow='->', arrowwidth=.2))
        self.anchors['base'] = (0, 0)
        self.anchors['collector'] = (bjt_emx, -bjt_emy-bjt_a)
        self.anchors['emitter'] = (bjt_emx, bjt_emy+bjt_a)


class BjtPnp2c(BjtPnp):
    ''' PNP Bipolar Junction Transistor with 2 collectors

        Args:
            circle: Draw circle around the transistor

        Anchors:
            * collector
            * emitter
            * base
            * C2
            * center
    '''
    def __init__(self, *, circle: Optional[bool] = None, **kwargs):
        super().__init__(circle=circle, **kwargs)
        bjt_2c_dy = -.25
        self.segments.append(Segment([(bjt_v, -bjt_a-bjt_2c_dy),
                                      (bjt_emx, -bjt_emy-bjt_2c_dy)]))
        self.anchors['C2'] = (bjt_emx, -bjt_emy-bjt_2c_dy)
        self.C2: Point


bjt_r = .55  # BJT circle radius
bjt_lead_dflt = 0.18  # Default lead length
bjt_diag_ofst = 0.2  # Distance from base to diagonals
bjt_base_w = .75  # Length of base bar
bjt_base_h = .42  # Height of base bar above leads
bjt_width = 1.1   # Distance between leads


class Bjt2(Element2Term):
    ''' Bipolar Junction Transistor (untyped) which extends
        collector/emitter leads to the desired length

        Args:
            circle: Draw circle around the transistor

        Anchors:
            * collector
            * emitter
            * base
    '''
    _element_defaults = {
        'circle': False,
        'lblloc': 'bottom',
        'lblofst': 0.1,
        'theta': 90
    }
    def __init__(self, *, circle: Optional[bool] = None, **kwargs):
        super().__init__(**kwargs)
        self.elmparams['ilabel'] = 'bottom'  # Draw current labels on this side
        self.segments.append(Segment(((0, 0),
                                      (bjt_width/2-bjt_diag_ofst, bjt_base_h),
                                      (bjt_width/2+bjt_diag_ofst, bjt_base_h),
                                      (bjt_width, 0))))  # Diagonals and lead ends
        self.segments.append(Segment(((bjt_width/2-bjt_base_w/2, bjt_base_h),
                                      (bjt_width/2+bjt_base_w/2, bjt_base_h))))  # Base bar
        self.segments.append(Segment(((bjt_width/2, bjt_base_h),
                                      (bjt_width/2, bjt_base_h+bjt_base_w/2))))  # Base lead
        if self.params['circle']:
            self.segments.append(SegmentCircle((bjt_width/2, bjt_r/2), bjt_r))
        self.anchors['base'] = (bjt_width/2, bjt_base_h+bjt_base_w/2)
        self.anchors['icollector'] = (bjt_width, 0)
        self.anchors['iemitter'] = (0, 0)

    def _place_anchors(self, start, end):
        ''' Allow positioning anchors relative to extended endpoints
            before the element is placed
        '''
        super()._place_anchors(start, end)
        self.anchors['emitter'] = self.anchors['start']
        self.anchors['collector'] = self.anchors['end']
        if self._userparams.get('reverse', False):
            self.anchors['emitter'] = self.anchors['end']
            self.anchors['collector'] = self.anchors['start']


class BjtNpn2(Bjt2):
    ''' NPN Bipolar Junction Transistor which extends
        collector/emitter leads to the desired length

        Args:
            circle: Draw circle around the transistor

        Anchors:
            * collector
            * emitter
            * base
    '''
    def __init__(self, *, circle: Optional[bool] = None, **kwargs):
        super().__init__(circle=circle, **kwargs)
        self.segments.append(Segment([(0, 0), (bjt_width/2-bjt_diag_ofst, bjt_base_h)],
                                     arrow='<-', arrowwidth=.2))


class BjtPnp2(Bjt2):
    ''' PNP Bipolar Junction Transistor which extends
        collector/emitter leads to the desired length

        Args:
            circle: Draw circle around the transistor

        Anchors:
            * collector
            * emitter
            * base
    '''
    def __init__(self, *, circle: Optional[bool] = None, **kwargs):
        super().__init__(circle=circle, **kwargs)
        self.segments.append(Segment([(bjt_width/2+bjt_diag_ofst, bjt_base_h),
                                      (bjt_width, 0)],
                                     arrow='<-', arrowwidth=.2))
        self.anchors['icollector'], self.anchors['iemitter'] = self.anchors['iemitter'], self.anchors['icollector']

    def _place_anchors(self, start, end):
        ''' Allow positioning anchors relative to extended endpoints
            before the element is placed
        '''
        super()._place_anchors(start, end)
        self.anchors['emitter'] = self.anchors['end']
        self.anchors['collector'] = self.anchors['start']
        if self._userparams.get('reverse', False):
            self.anchors['emitter'] = self.anchors['start']
            self.anchors['collector'] = self.anchors['end']


class BjtPnp2c2(BjtPnp2):
    ''' 2-Collector PNP Bipolar Junction Transistor which extends
        collector/emitter leads to the desired length

        Args:
            circle: Draw circle around the transistor

        Anchors:
            * collector
            * emitter
            * base
            * C2
    '''
    def __init__(self, *, circle: Optional[bool] = None, **kwargs):
        super().__init__(circle=circle, **kwargs)
        bjt_2c_dy = .25
        self.segments.append(Segment([(bjt_2c_dy, 0),
                                      (bjt_2c_dy+bjt_width/2-bjt_diag_ofst, bjt_base_h)]))
        self.anchors['C2'] = (bjt_2c_dy, 0)
        self.C2: Point



from .elements import Element, ElementDrawing, Element2Term
from .container import Container
from .twoterm import (Resistor, ResistorIEEE, ResistorIEC, ResistorVar, ResistorVarIEEE,
                      ResistorVarIEC, Thermistor, Photoresistor, PhotoresistorIEEE, PhotoresistorIEC,
                      Rshunt, Capacitor, Capacitor2, CapacitorVar, CapacitorTrim, Diode, Schottky, DiodeTunnel,
                      DiodeShockley, Zener, Varactor, LED, LED2, Photodiode, Potentiometer, PotentiometerIEEE,
                      PotentiometerIEC, Diac, Triac, SCR, Memristor, Memristor2, Josephson, Fuse, FuseUS, FuseIEEE,
                      FuseIEC, Inductor, Inductor2, Crystal, Breaker, CPE, SparkGap, RBox, RBoxVar,
                      PotBox, PhotoresistorBox, Nullator, Norator, CurrentMirror, VoltageMirror)
from .oneterm import Ground, GroundSignal, GroundChassis, Antenna, AntennaLoop, AntennaLoop2, Vss, Vdd, NoConnect
from .opamp import Opamp
from .sources import (Source, SourceV, SourceI, SourceSin, SourcePulse, SourceSquare, SourceTriangle, SourceRamp,
                      SourceControlled, SourceControlledV, SourceControlledI, BatteryCell, Battery, MeterV, MeterI,
                      MeterA, MeterOhm, Lamp, Lamp2, Solar, Neon)
from .switches import (Switch, SwitchSpdt, SwitchSpdt2, SwitchDpst, SwitchDpdt, Button, SwitchReed,
                       SwitchRotary, SwitchDIP)
from .transistors import (NFet, PFet, AnalogNFet, AnalogPFet, AnalogBiasedFet,
                          JFet, JFetN, JFetP, Bjt, BjtNpn, BjtPnp, BjtPnp2c, Bjt2, BjtNpn2, BjtPnp2,
                          BjtPnp2c2, NFet2, PFet2, JFet2, JFetN2, JFetP2, NMos, PMos, NMos2, PMos2)
from .misc import Speaker, Mic, Motor, AudioJack
from .xform import Transformer
from .cables import Coax, Triax
from .intcircuits import (IcPin, Ic, Multiplexer, IcDIP, VoltageRegulator, DFlipFlop, JKFlipFlop, Ic555,
                          SevenSegment, sevensegdigit)
from .lines import (Line, DataBusLine, Dot, Arrowhead, Arrow, DotDotDot, Wire, Gap, Label, Tag, CurrentLabel,
                    CurrentLabelInline, ZLabel, LoopCurrent, LoopArrow, Rect, Arc2, Arc3, ArcZ, ArcN, ArcLoop,
                    Annotate, Encircle, EncircleBox)
from .connectors import OrthoLines, RightLines, Header, Jumper, BusConnect, BusLine, DB25, DB9, CoaxConnect, Plug, Jack, Terminal
from .compound import ElementCompound, Optocoupler, Relay, Rectifier, Wheatstone
from .twoports import (ElementTwoport, TwoPort, VoltageTransactor, TransimpedanceTransactor, CurrentTransactor,
                       TransadmittanceTransactor, Nullor, VMCMPair)
from .outlets import (OutletA, OutletB, OutletC, OutletD, OutletE, OutletF, OutletG, OutletH, OutletI, OutletJ,
                      OutletK, OutletL)
from .image import ElementImage


__all__ = [
    "Element", "ElementDrawing", "ElementImage", "Element2Term", "Container",
    "Resistor", "ResistorIEEE", "ResistorIEC", "ResistorVar", "ResistorVarIEEE", "ResistorVarIEC", "Rshunt",
    "Thermistor", "Photoresistor", "PhotoresistorIEEE", "PhotoresistorIEC", "Capacitor", "Capacitor2",
    "CapacitorVar", "CapacitorTrim", "Diode", "Schottky", "DiodeTunnel", "DiodeShockley", "Zener", "Varactor",
    "LED", "LED2", "Photodiode", "Potentiometer", "PotentiometerIEEE", "PotentiometerIEC", "Diac", "Triac", "SCR",
    "Memristor", "Memristor2", "Josephson", "Fuse", "FuseUS", "FuseIEEE", "FuseIEC", "Inductor", "Inductor2",
    "Crystal", "Breaker", "ResistorVarIEC", "CPE", "SparkGap", "RBox", "RBoxVar", "PotBox", "PhotoresistorBox",
    "Nullator", "Norator", "CurrentMirror", "VoltageMirror",
    "Ground", "GroundSignal", "GroundChassis", "Antenna", "AntennaLoop", "AntennaLoop2", "Vss", "Vdd", "NoConnect",
    "Opamp", "Source", "SourceV", "SourceI", "SourceSin", "SourcePulse", "SourceSquare", "SourceTriangle",
    "SourceRamp", "SourceControlled", "SourceControlledV", "SourceControlledI", "BatteryCell", "Battery", "MeterV",
    "MeterI", "MeterA", "MeterOhm", "Lamp", "Lamp2", "Solar", "Neon", "Switch", "SwitchSpdt", "SwitchSpdt2", "SwitchDpst",
    "SwitchDpdt", "Button", "SwitchReed", "SwitchRotary", "SwitchDIP", "NFet", "PFet", "AnalogNFet", "AnalogPFet",
    "AnalogBiasedFet", "JFet", "JFetN", "JFetP",
    "Bjt", "BjtNpn", "BjtPnp", "BjtPnp2c", "Bjt2", "BjtNpn2", "BjtPnp2", "BjtPnp2c2", "NFet2", "PFet2", "JFet2",
    "JFetN2", "JFetP2", "NMos", "PMos", "NMos2", "PMos2",
    "Speaker", "Mic", "Motor", "AudioJack", "Transformer", "Coax", "Triax",
    "IcPin", "Ic", "Multiplexer", "IcDIP", "VoltageRegulator", "DFlipFlop", "JKFlipFlop", "Ic555", "SevenSegment",
    "sevensegdigit", "Line", "DataBusLine", "Dot", "Annotate", "ZLabel", "Arrowhead", "Arrow", "DotDotDot", "Wire",
    "Gap", "Label", "Tag", "CurrentLabel", "CurrentLabelInline", "LoopCurrent", "LoopArrow", "Rect",
    "Arc2", "Arc3", "ArcZ", "ArcN", "ArcLoop", "Encircle", "EncircleBox", "OrthoLines", "RightLines", "Header",
    "Jumper", "BusConnect", "BusLine", "DB25", "DB9", "CoaxConnect", "Plug", "Jack", "Terminal", "ElementCompound", "Optocoupler",
    "Relay", "Rectifier", "Wheatstone", "ElementTwoport", "TwoPort", "VoltageTransactor", "TransimpedanceTransactor",
    "CurrentTransactor", "TransadmittanceTransactor", "Nullor", "VMCMPair", "OutletA", "OutletB", "OutletC", "OutletD",
    "OutletE", "OutletF", "OutletG", "OutletH", "OutletI", "OutletJ", "OutletK", "OutletL"]


STYLE_IEEE = {'Resistor': ResistorIEEE,
              'ResistorVar': ResistorVarIEEE,
              'Potentiometer': PotentiometerIEEE,
              'Photoresistor': PhotoresistorIEEE,
              'Fuse': FuseUS}
STYLE_IEC = {'Resistor': ResistorIEC,
             'ResistorVar': ResistorVarIEC,
             'Potentiometer': PotentiometerIEC,
             'Photoresistor': PhotoresistorIEC,
             'Fuse': FuseIEC}


def style(style):
    ''' Set global element style

        Args:
            style: dictionary of elementname: Element
            to change the element module namespace.
            Use `elements.STYLE_IEEE` or `elements.STYLE_IEC`
            to define U.S./IEEE or European/IEC element styles.
    '''
    for name, element in style.items():
        globals()[name] = element



from .logic import And, Nand, Or, Nor, Xor, Xnor, Buf, Not, NotNot, Tristate, Tgate, Schmitt, SchmittNot, SchmittAnd, SchmittNand
from .kmap import Kmap
from .table import Table
from .timing import TimingDiagram
from ..elements import Arrow, Arrowhead, Dot, Line, Wire, Arc2, Arc3, ArcLoop

__all__ = ['And', 'Nand', 'Or', 'Nor', 'Xor', 'Xnor', 'Buf', 'Not', 'NotNot', 'Tristate', 'Tgate', 'Schmitt', 'SchmittNot',
           'SchmittAnd', 'SchmittNand', 'Kmap', 'Table', 'TimingDiagram', 'Arrow', 'Arrowhead', 'Dot', 'Line',
           'Wire', 'Arc2', 'Arc3', 'ArcLoop']



''' Container Element '''
from __future__ import annotations
import math
from typing import Union, Optional, TYPE_CHECKING

from .elements import Element
from ..segments import Segment, SegmentPoly, BBox
from .. import drawing_stack

if TYPE_CHECKING:
    from ..schemdraw import Drawing


class Container(Element):
    ''' Container element - a group of elements surrounded by a box
    
        Args:
            drawing: The schemdraw Drawing or another Container to contain
            cornerradius: Radius to round the corners of the box [default: 0.3]
            padx: Spacing from element to box in x direction [default: 0.75]
            pady: Spacing from element to box in y direction [default: 0.75]
    '''
    _element_defaults = {
        'cornerradius': 0.3,
        'padx': 0.75,
        'pady': 0.75,
        'lblloc': 'NW',
        'lblofst': (.1, -.1),
        'lblalign': ('left', 'top'),
        'theta': 0,
        'anchor': 'boxNW'
    }
    def __init__(self,
                 drawing: Union['Drawing', 'Container'],
                 *,
                 cornerradius: Optional[float] = None,
                 padx: Optional[float] = None,
                 pady: Optional[float] = None):
        super().__init__()
        self.drawing = drawing
        self.elements: list[Element] = []

    def container(self,
                  cornerradius: Optional[float] = None,
                  padx: Optional[float] = None,
                  pady: Optional[float] = None) -> 'Container':
        ''' Add a container to the container. Use as a context manager,
            such that elemnents inside the `with` are surrounded by
            the container.
            
            >>> with container.container():
            >>>    elm.Resistor()
            >>>    ...

            Args:
            cornerradius: Radius to round the corners of the box [default: 0.3]
            padx: Spacing from element to box in x direction [default: 0.75]
            pady: Spacing from element to box in y direction [default: 0.75]
        '''
        return Container(self, cornerradius=cornerradius, padx=padx, pady=pady)

    def __iadd__(self, element: Element) -> 'Container':
        return self.add(element)

    def add(self, element: Element) -> 'Container':
        ''' Add an element to the container '''
        self.elements.append(element)
        self.drawing.add(element)
        return self

    def __enter__(self):
        drawing_stack.push_drawing(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        drawing_stack.push_element(self)
        drawing_stack.pop_drawing(self)

    def __contains__(self, element):
        return element in self.elements

    def container_bbox(self, transform: bool = True) -> BBox:
        ''' Bounding box of the contents only '''
        xmin = math.inf
        xmax = -math.inf
        ymin = math.inf
        ymax = -math.inf
        for element in self.elements:
            bbox = element.get_bbox(transform=transform)
            xmin = min(bbox.xmin, xmin)
            xmax = max(bbox.xmax, xmax)
            ymin = min(bbox.ymin, ymin)
            ymax = max(bbox.ymax, ymax)
        return BBox(xmin, ymin, xmax, ymax)

    def get_bbox(self, transform: bool = False, includetext: bool = True) -> BBox:
        ''' Bounding box including border '''
        bbox = self.container_bbox(transform=transform)
        return BBox(bbox.xmin-self.params['padx'],
                    bbox.ymin-self.params['pady'],
                    bbox.xmax+self.params['padx'],
                    bbox.ymax+self.params['pady'])

    def _place(self, dwgxy, dwgtheta, **dwgparams):
        bbox = self.container_bbox()
        w = bbox.xmax - bbox.xmin + self.params['padx']*2
        h = bbox.ymax - bbox.ymin + self.params['pady']*2
        cornerradius = self.params['cornerradius']
        bbox = BBox(bbox.xmin-self.params['padx'],
                    bbox.ymin-self.params['pady'],
                    bbox.xmax+self.params['padx'],
                    bbox.ymax+self.params['pady'])
        self.at((bbox.xmin, bbox.ymax))
        if cornerradius > 0:
            self.segments = [SegmentPoly(
                [(0, h/2), (w, h/2), (w, -h/2), (0, -h/2)],
                 cornerradius=cornerradius)]
        else:
            self.segments.append(Segment([(0, 0), (0, h/2), (w, h/2),
                                          (w, -h/2), (0, -h/2), (0, 0)]))
        self.anchors['center'] = (w/2, 0)
        self.anchors['boxNW'] = (0, h/2)
        self.anchors['N'] = (w/2, h/2)
        self.anchors['E'] = (w, 0)
        self.anchors['S'] = (w/2, -h/2)
        self.anchors['W'] = (0, 0)
        k = cornerradius - cornerradius*math.sqrt(2)/2
        self.anchors['NE'] = (w-k, h/2-k)
        self.anchors['NW'] = (k, h/2-k)
        self.anchors['SE'] = (w-k, -h/2+k)
        self.anchors['SW'] = (k, -h/2+k)
        self.anchors['NNE'] = (3*w/4 if cornerradius < w/4 else w-cornerradius, h/2)
        self.anchors['NNW'] = (w/4 if cornerradius < w/4 else cornerradius, h/2)
        self.anchors['SSE'] = (3*w/4 if cornerradius < w/4 else w-cornerradius, -h/2)
        self.anchors['SSW'] = (w/4 if cornerradius < w/4 else cornerradius, -h/2)
        self.anchors['ENE'] = (w, h/4 if cornerradius < h/4 else h/2-cornerradius)
        self.anchors['ESE'] = (w, -h/4 if cornerradius < h/4 else -h/2+cornerradius)
        self.anchors['WNW'] = (0, h/4 if cornerradius < h/4 else h/2-cornerradius)
        self.anchors['WSW'] = (0, -h/4 if cornerradius < h/4 else -h/2+cornerradius)
        return super()._place(dwgxy, dwgtheta, **dwgparams)



''' Markdown table.

    Put in the logic module since logic truth tables were the first
    use case for tables.
'''
from __future__ import annotations

import re

from ..segments import Segment, SegmentText
from ..elements import Element
from ..backends import svg
from typing import Optional


def parse_colfmt(colfmt: str) -> tuple[str, str]:
    ''' Parse the column formatter string, using LaTeX style table formatter
        (e.g. "cc|c")

        Args:
            colfmt: column format string

        Returns:
            justifications: string of all justification characters, length = n
            bars: string of all separator characters, |, ǁ, or ., length = n+1
    '''
    colfmt = re.sub(r'\|\|', 'ǁ', colfmt)
    while True:
        out = re.sub(r'([lcr])([lcr])', r'\1.\2', colfmt)
        if out == colfmt:
            break
        colfmt = out

    if not colfmt.startswith('|') or colfmt.startswith('ǁ'):
        colfmt = '.' + colfmt
    if not colfmt.endswith('|') or colfmt.endswith('ǁ'):
        colfmt += '.'

    bars = colfmt[::2]
    justs = colfmt[1::2]
    return justs, bars


class Table(Element):
    r''' Table Element for drawing rudimentary Markdown formatted tables, such
        as logic truth tables.

        Args:
            table: Table definition, as markdown string. Columns separated by \|.
                Separator rows contain --- or === between column separators.
            colfmt: Justification and vertical separators to draw for each column,
                similar to LaTeX tabular environment parameter. Justification
                characters include 'c', 'r', and 'l' for center, left, and right
                justification. Separator characters may be '\|' for a single
                vertical bar, or '\|\|' or 'ǁ' for a double vertical bar, or omitted
                for no bar. Example: 'cc|c'.
            fontsize: Point size of table font
            font: Name of table font

        Example Table:
            \| A \| B \| Y \|
            \|---\|---\|---\|
            \| 0 \| 0 \| 1 \|
            \| 0 \| 1 \| 0 \|
            \| 1 \| 0 \| 0 \|
            \| 1 \| 1 \| 0 \|
    '''
    def __init__(self, table: str, colfmt: Optional[str] = None,
                 fontsize: float = 12, font: str = 'sans', **kwargs):
        super().__init__(**kwargs)
        self.table = table

        doublebarsep = .03
        rowpad = .3
        colpad = .4
        kwargs.setdefault('lw', 1)

        rows = self.table.strip().splitlines()
        rowfmt = ''
        for row in rows:
            if '---' in row:
                rowfmt += '|'
            elif '===' in row:
                rowfmt += 'ǁ'
            else:
                rowfmt += 'l'
        rowjusts, rowbars = parse_colfmt(rowfmt)

        if colfmt in [None, '']:
            colfmt = 'c' * len(rows[0].strip('| ').split('|'))
        coljusts, colbars = parse_colfmt(colfmt)  # type: ignore

        ncols = len(coljusts)
        nrows = len(rowjusts)
        rows = [r for r in rows if '---' not in r and '===' not in r]
        if len(rows[0].strip('| ').split('|')) != ncols:
            raise ValueError('Number of columns in table does not match number of columns in colfmt string.')

        colwidths = [0.] * ncols
        rowheights = [0.] * nrows
        for k, row in enumerate(rows):
            cells = [c.strip() for c in row.strip('| ').split('|')]
            for i, cell in enumerate(cells):
                txtw, txth, _ = svg.text_size(cell, font=font, size=fontsize)
                colwidths[i] = max(colwidths[i], txtw/72*2+colpad)
                rowheights[k] = max(rowheights[k], txth/72*2+rowpad)

        totheight = sum(rowheights)
        totwidth = sum(colwidths)

        # Frame
        for i, colbar in enumerate(colbars):
            colx = sum(colwidths[:i])
            if colbar == '|':
                self.segments.append(Segment([(colx, 0), (colx, -totheight)], **kwargs))
            elif colbar == 'ǁ':
                self.segments.append(Segment([(colx-doublebarsep, 0), (colx-doublebarsep, -totheight)], **kwargs))
                self.segments.append(Segment([(colx+doublebarsep, 0), (colx+doublebarsep, -totheight)], **kwargs))
        for i, rowbar in enumerate(rowbars):
            rowy = -sum(rowheights[:i])
            if rowbar == '|':
                self.segments.append(Segment([(0, rowy), (totwidth, rowy)], **kwargs))
            elif rowbar == 'ǁ':
                self.segments.append(Segment([(0, rowy-doublebarsep), (totwidth, rowy-doublebarsep)], **kwargs))
                self.segments.append(Segment([(0, rowy+doublebarsep), (totwidth, rowy+doublebarsep)], **kwargs))

        # Text
        for r, row in enumerate(rows):
            cells = [c.strip() for c in row.strip('| ').split('|')]
            for c, cell in enumerate(cells):
                cellx = sum(colwidths[:c])
                celly = -sum(rowheights[:r]) - rowheights[r]
                halign = {'c': 'center', 'l': 'left', 'r': 'right'}.get(coljusts[c])
                if halign == 'center':
                    cellx += colwidths[c]/2
                elif halign == 'right':
                    cellx += colwidths[c] - colpad/2
                else:
                    cellx += colpad/2
                self.segments.append(
                    SegmentText((cellx, celly), cell, font=font, fontsize=fontsize,
                                align=(halign, 'bottom')))    # type: ignore



from .dsp import (Square, Circle, Sum, SumSigma, Mixer, Speaker, Amp, OscillatorBox, Oscillator, Filter, Adc,
                  Dac, Demod, Circulator, Isolator, VGA)
from ..elements import Arrowhead, Antenna, Dot, Arrow, Line, Ic, IcPin, Multiplexer, Wire
from ..flow import Box

__all__ = ['Square', 'Circle', 'Sum', 'SumSigma', 'Mixer', 'Speaker', 'Amp', 'OscillatorBox', 'Oscillator',
           'Filter', 'Adc', 'Dac', 'Demod', 'Circulator', 'Isolator', 'VGA', 'Arrowhead', 'Antenna', 'Dot',
           'Arrow', 'Line', 'Ic', 'IcPin', 'Multiplexer', 'Wire', 'Box']



''' Operation amplifier '''
from __future__ import annotations
from typing import Optional
import math

from .elements import Element, gap
from ..segments import Segment


oa_back = 2.5
oa_xlen = oa_back * math.sqrt(3)/2
oa_lblx = oa_xlen/8
oa_pluslen = .2


class Opamp(Element):
    ''' Operational Amplifier.

        Keyword Args:
            sign: Draw +/- labels at each input
            leads: Draw short leads on input/output

        Anchors:
            * in1
            * in2
            * out
            * vd
            * vs
            * n1
            * n2
            * n1a
            * n2a
    '''
    _element_defaults = {
        'sign': True,
        'leads': False
    }
    def __init__(self, *,
                 sign: Optional[bool] = None,
                 leads: Optional[bool] = None,
                 **kwargs):
        super().__init__(**kwargs)
        leadlen = oa_back/4
        x = 0 if not self.params['leads'] else leadlen

        self.segments.append(Segment(
            [(x, 0), (x, oa_back/2), (x+oa_xlen, 0), (x, -oa_back/2), (x, 0),
             gap, (x+oa_xlen, 0)]))

        if self.params['sign']:
            self.segments.append(Segment(
                [(x+oa_lblx-oa_pluslen/2, oa_back/4),
                 (x+oa_lblx+oa_pluslen/2, oa_back/4)]))
            self.segments.append(Segment(
                [(x+oa_lblx-oa_pluslen/2, -oa_back/4),
                 (x+oa_lblx+oa_pluslen/2, -oa_back/4)]))
            self.segments.append(Segment(
                [(x+oa_lblx, -oa_back/4-oa_pluslen/2),
                 (x+oa_lblx, -oa_back/4+oa_pluslen/2)]))

        if self.params['leads']:
            self.segments.append(Segment([(0, oa_back/4), (leadlen, oa_back/4)]))
            self.segments.append(Segment([(0, -oa_back/4), (leadlen, -oa_back/4)]))
            self.segments.append(Segment([(leadlen+oa_xlen, 0), (oa_xlen + 2*leadlen, 0)]))

        self.anchors['out'] = (oa_xlen+2*x, 0)
        self.anchors['in1'] = (0, oa_back/4)
        self.anchors['in2'] = (0, -oa_back/4)
        self.anchors['center'] = (x+oa_xlen/2, 0)
        self.anchors['vd'] = (x+oa_xlen/3, .84)
        self.anchors['vs'] = (x+oa_xlen/3, -.84)
        self.anchors['n1'] = (x+oa_xlen*2/3, -.42)
        self.anchors['n2'] = (x+oa_xlen*2/3, .42)
        self.anchors['n1a'] = (x+oa_xlen*.9, -.13)
        self.anchors['n2a'] = (x+oa_xlen*.9, .13)
        self.elmparams['drop'] = (2*x+oa_xlen, 0)



''' Schemdraw Drawing class '''

from __future__ import annotations
from typing import Any, MutableMapping, Union, Optional, TYPE_CHECKING
from collections import ChainMap
import warnings
import math

from . import default_canvas
from .types import BBox, Backends, ImageFormat, Linestyle, XY, ImageType
from .elements import Element, Container
from .segments import SegmentType
from .util import Point
from .backends.svg import Figure as svgFigure
from . import drawing_stack

if TYPE_CHECKING:
    import xml.etree.ElementTree.Element  # type: ignore

try:
    from .backends.mpl import Figure as mplFigure
    default_canvas.default_canvas = 'matplotlib'
    if TYPE_CHECKING:
        import matplotlib.pyplot.Axes   # type: ignore
except ImportError:
    mplFigure = None  # type: ignore
    default_canvas.default_canvas = 'svg'


def use(backend: Backends = 'matplotlib') -> None:
    ''' Change default backend, either 'matplotlib' or 'svg' '''
    if backend == 'matplotlib':
        if mplFigure is None:
            raise ValueError('Could not import Matplotlib.')
    default_canvas.default_canvas = backend


def config(unit: float = 3.0, inches_per_unit: float = 0.5,
           lblofst: float = 0.1, fontsize: float = 14.,
           font: str = 'sans-serif', color: str = 'black',
           lw: float = 2., ls: Linestyle = '-',
           fill: Optional[str] = None, bgcolor: Optional[str] = None,
           margin: float = 0.1, mathfont: Optional[str] = None) -> None:
    ''' Set global schemdraw style configuration

        Args:
            unit: Full length of a 2-terminal element. Inner zig-zag portion
                of a resistor is 1.0 units.
            inches_per_unit: Inches per drawing unit for setting drawing scale
            lblofst: Default offset between element and its label
            fontsize: Default font size for text labels
            font: Default font family for text labels
            color: Default color name or RGB (0-1) tuple
            lw: Default line width for elements
            ls: Default line style
            fill: Deault fill color for closed elements
            margin: White space around the drawing in drawing units
            mathont: Font for math delimited by $..$
    '''
    schemdrawstyle['unit'] = unit
    schemdrawstyle['inches_per_unit'] = inches_per_unit
    schemdrawstyle['lblofst'] = lblofst
    schemdrawstyle['fontsize'] = fontsize
    schemdrawstyle['font'] = font
    schemdrawstyle['color'] = color
    schemdrawstyle['lw'] = lw
    schemdrawstyle['ls'] = ls
    schemdrawstyle['fill'] = fill
    schemdrawstyle['margin'] = margin
    if bgcolor:
        schemdrawstyle['bgcolor'] = bgcolor
    if mathfont:
        schemdrawstyle['mathfont'] = mathfont


def debug(dwgbbox: bool = True,
          elmbbox: bool = True):
    ''' Debug - draw element and/or drawing bounding boxes '''
    schemdrawstyle['dwgbbox'] = dwgbbox
    schemdrawstyle['elmbbox'] = elmbbox


schemdrawstyle: dict[str, Any] = {}  # Global style
config()  # Initialize default configuration


def theme(theme='default'):
    ''' Set schemdraw theme (line color and background color).
        Themes match those in jupyter-themes package
        (https://github.com/dunovank/jupyter-themes).

        Available themes:
            * default (black on white)
            * dark (white on black)
            * solarizedd
            * solarizedl
            * onedork
            * oceans16
            * monokai
            * gruvboxl
            * gruvboxd
            * grade3
            * chesterish
    '''
    if theme == 'default':
        config(bgcolor='white')
    elif theme == 'dark':
        schemdrawstyle['color'] = 'white'
        schemdrawstyle['bgcolor'] = 'black'
    elif theme == 'solarizedd':
        schemdrawstyle['bgcolor'] = '#002b36'
        schemdrawstyle['color'] = '#657b83'
    elif theme == 'solarizedl':
        schemdrawstyle['bgcolor'] = '#eee8d5'
        schemdrawstyle['color'] = '#073642'
    elif theme == 'onedork':
        schemdrawstyle['bgcolor'] = '#373e4b'
        schemdrawstyle['color'] = '#899ab8'
    elif theme == 'oceans16':
        schemdrawstyle['bgcolor'] = '#384151'
        schemdrawstyle['color'] = '#CDD2E9'
    elif theme == 'monokai':
        schemdrawstyle['bgcolor'] = '#232323'
        schemdrawstyle['color'] = '#BBBBBB'
    elif theme == 'gruvboxl':
        schemdrawstyle['bgcolor'] = '#ebdbb2'
        schemdrawstyle['color'] = '#3c3836'
    elif theme == 'gruvboxd':
        schemdrawstyle['bgcolor'] = '#1d2021'
        schemdrawstyle['color'] = '#d5c4a1'
    elif theme == 'grade3':
        schemdrawstyle['bgcolor'] = '#ffffff'
        schemdrawstyle['color'] = '#3f3d46'
    elif theme == 'chesterish':
        schemdrawstyle['bgcolor'] = '#323A48'
        schemdrawstyle['color'] = '#92A2BD'
    else:
        raise ValueError(f'Unknown theme {theme}')


class Drawing:
    ''' A schematic drawing

        See `schemdraw.config` method for argument defaults

        Args:
            canvas: Canvas to draw on when using Drawing context manager.
                Can be string 'matplotlib' or 'svg' to create new canvas
                with these backends, or an instance of a matplotlib axis,
                or an instance of xml.etree.ElementTree containing SVG.
                Default is value set by schemdraw.use().
            file: optional filename to save on exiting context manager
                or calling draw method.
            show: Show the drawing after exiting context manager

        Attributes:
            here: (xy tuple) Current drawing position. The next element will
                be added at this position unless specified otherwise.
            theta: (float) Current drawing angle, in degrees. The next
                element will be added with this angle unless specified
                otherwise.
    '''     
    def __init__(self, canvas: Optional[Union[Backends,
                                        xml.etree.ElementTree.Element,
                                        matplotlib.pyplot.Axes]] = None,
                 file: Optional[str] = None, show: bool = True, **kwargs):
        self.outfile = file
        self.canvas = canvas
        self.show = show
        self.elements: list[Element] = []
        self.anchors: MutableMapping[str, Union[Point, tuple[float, float]]] = {}  # Untransformed anchors

        if 'backend' in kwargs:
            self.canvas = kwargs.pop('backend')
            warnings.warn('Use of `backend` is deprecated. Use `canvas`.',
                          DeprecationWarning, stacklevel=2)

        self.dwgparams: dict[str, Any] = schemdrawstyle.copy()
        self.dwgparams.update(kwargs)  # To maintain support for arguments that moved to config method
        self.unit = kwargs.get('unit', schemdrawstyle.get('unit'))

        self._here: XY = Point((0, 0))
        self._theta: float = 0
        self._state: list[tuple[Point, float]] = []  # Push/Pop stack
        self._interactive = False
        self.fig: Optional[Union[mplFigure, svgFigure]] = None

    @property
    def here(self):
        drawing_stack.push_element(None)
        return self._here

    @here.setter
    def here(self, value):
        drawing_stack.push_element(None)
        self._here = value

    @property
    def theta(self):
        drawing_stack.push_element(None)
        return self._theta

    @theta.setter
    def theta(self, value):
        self._theta = value

    def container(self, cornerradius: float = .3,
                  padx: float = .75, pady: float = .75):
        ''' Add a container to the drawing. Use as a context manager,
            such that elemnents inside the `with` are surrounded by
            the container.
            
            >>> with drawing.container():
            >>>    elm.Resistor()
            >>>    ...

            Args:
                cornerradius: radius for box corners
                padx: space between contents and border in x direction
                pady: space between contents and border in y direction
        '''
        return Container(self, cornerradius=cornerradius, padx=padx, pady=pady)

    def __enter__(self):
        drawing_stack.push_drawing(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        ''' Exit context manager - save to file and display '''
        drawing_stack.push_element(None)
        drawing_stack.pop_drawing(self)

        if self.outfile is not None:
            self.save(self.outfile)
        if not self.fig:
            self.draw(show=False)
        if self.show and not hasattr(self.canvas, 'plot'):
            try:
                display(self.fig)
            except NameError:  # Not in Jupyter/IPython
                self.fig.show()

    def __getattr__(self, name: str) -> Any:
        ''' Allow getting anchor position as attribute '''
        if name in vars(self).get('anchors', {}):
            return vars(self).get('anchors')[name]  # type: ignore
        raise AttributeError(f"'Drawing' has no attribute {name}")

    def __contains__(self, element):
        return element in self.elements

    def interactive(self, interactive: bool = True):
        ''' Enable interactive mode (matplotlib backend only). Matplotlib
            must also be set to interactive with `plt.ion()`.
        '''
        self._interactive = interactive

    def get_bbox(self) -> BBox:
        ''' Get drawing bounding box '''
        xmin = math.inf
        xmax = -math.inf
        ymin = math.inf
        ymax = -math.inf
        for element in self.elements:
            bbox = element.get_bbox(transform=True)
            xmin = min(bbox.xmin, xmin)
            xmax = max(bbox.xmax, xmax)
            ymin = min(bbox.ymin, ymin)
            ymax = max(bbox.ymax, ymax)
        return BBox(xmin, ymin, xmax, ymax)

    def get_segments(self) -> list[SegmentType]:
        ''' Get flattened list of all segments in the drawing '''
        segments = []
        for element in self.elements:
            # Exclude drawing params from the chain
            params = ChainMap(element._userparams, element.elmparams, element.defaults)
            segments.extend([s.xform(element.transform, **params)
                             for s in element.segments])
        return segments

    def _repr_svg_(self):
        ''' SVG representation for Jupyter '''
        return self.draw(show=False).getimage('svg').decode()

    def _repr_png_(self):
        ''' PNG representation for Jupyter '''
        if self.canvas == 'matplotlib' or hasattr(self.canvas, 'plot'):
            return self.draw(show=False).getimage('png')
        return None

    def __iadd__(self, element: Element):
        ''' In-place add element, via += operator. '''
        self.add(element)
        return self

    def add(self, element: Element) -> Element:
        ''' Add an element to the drawing.

            Args:
                element: The element to add.
        '''
        self._here, self._theta = element._place(self._here, self._theta, **self.dwgparams)
        self.elements.append(element)

        if self._interactive:
            if self.fig is None:
                self.fig = mplFigure(
                    inches_per_unit=self.dwgparams.get('inches_per_unit'))
                if 'bgcolor' in self.dwgparams:
                    self.fig.bgcolor(self.dwgparams['bgcolor'])
            element._draw(self.fig)
            self.fig.set_bbox(self.get_bbox())  # type: ignore
            self.fig.getimage()  # type: ignore
        else:
            self.fig = None  # Clear any existing figure
        return element

    def add_elements(self, *elements: Element) -> None:
        ''' Add multiple elements to the drawing '''
        for element in elements:
            self.add(element)

    def undo(self) -> None:
        ''' Removes previously added element '''
        self.elements.pop(-1)
        self.fig.clear()  # type: ignore
        for element in self.elements:
            element._draw(self.fig)
        self._here, self._theta = self.elements[-1].absdrop
        self.fig.set_bbox(self.get_bbox())  # type: ignore
        self.fig.getimage()  # type: ignore

    def move(self, dx: float = 0, dy: float = 0) -> None:
        ''' Move the current drawing position

            Args:
                dx: change in x position
                dy: change in y position
        '''
        drawing_stack.push_element(None)
        self._here = Point((self._here[0] + dx, self._here[1] + dy))

    def move_from(self, ref: Point, dx: float = 0, dy: float = 0, theta: Optional[float] = None) -> None:
        ''' Move drawing position relative to the reference point. Change drawing
            theta if provided.
        '''
        drawing_stack.push_element(None)
        self._here = (ref.x + dx, ref.y + dy)
        if theta is not None:
            self._theta = theta

    def set_anchor(self, name: str) -> None:
        ''' Define a Drawing anchor at the current drawing position '''
        self.anchors[name] = self._here

    def push(self) -> None:
        ''' Push/save the drawing state.
            Drawing.here and Drawing.theta are saved.
        '''
        drawing_stack.push_element(None)
        self._state.append((Point(self._here), self._theta))

    def pop(self) -> None:
        ''' Pop/load the drawing state. Location and angle are returned to
            previously pushed state.
        '''
        drawing_stack.push_element(None)
        if len(self._state) > 0:
            self._here, self._theta = self._state.pop()

    def config(self, unit: Optional[float] = None, inches_per_unit: Optional[float] = None,
               fontsize: Optional[float] = None, font: Optional[str] = None,
               color: Optional[str] = None, lw: Optional[float] = None, ls: Optional[Linestyle] = None,
               fill: Optional[str] = None, bgcolor: Optional[str] = None,
               margin: Optional[float] = None,
               mathfont: Optional[str] = None) -> None:
        ''' Set Drawing configuration, overriding schemdraw global config.

            Args:
                unit: Full length of a 2-terminal element. Inner zig-zag portion
                    of a resistor is 1.0 units.
                inches_per_unit: Inches per drawing unit for setting drawing scale
                fontsize: Default font size for text labels
                font: Default font family for text labels
                mathfont: Default font family or filename math text delimited by $..$
                color: Default color name or RGB (0-1) tuple
                lw: Default line width for elements
                ls: Default line style
                fill: Deault fill color for closed elements
                margin: White space around the drawing in drawing units
        '''
        if unit is not None:
            self.unit = unit
            self.dwgparams['unit'] = unit
        if inches_per_unit is not None:
            self.dwgparams['inches_per_unit'] = inches_per_unit
        if font is not None:
            self.dwgparams['font'] = font
        if fontsize is not None:
            self.dwgparams['fontsize'] = fontsize
        if color is not None:
            self.dwgparams['color'] = color
        if lw is not None:
            self.dwgparams['lw'] = lw
        if ls is not None:
            self.dwgparams['ls'] = ls
        if fill is not None:
            self.dwgparams['fill'] = fill
        if bgcolor is not None:
            self.dwgparams['bgcolor'] = bgcolor
        if margin is not None:
            self.dwgparams['margin'] = margin
        if mathfont is not None:
            self.dwgparams['mathfont'] = mathfont

    def _drawelements(self):
        ''' Draw all the elements on self.fig '''
        for element in self.elements:
            element._draw(self.fig)

    def _drawmpl(self, ax=None):
        ''' Draw on Matplotlib Axis '''
        if self.fig is None or ax is not None:
            self.fig = mplFigure(ax=ax,
                                 inches_per_unit=self.dwgparams.get('inches_per_unit'),
                                 margin=self.dwgparams['margin'],
                                 showbbox=self.dwgparams.get('dwgbbox', False))
            if 'bgcolor' in self.dwgparams:
                self.fig.bgcolor(self.dwgparams['bgcolor'])
        self.fig.set_bbox(self.get_bbox())  # type: ignore
        self._drawelements()

    def _drawsvg(self, svg=None):
        ''' Draw on SVG canvas '''
        if self.fig is None or svg is not None:
            self.fig = svgFigure(svg=svg, bbox=self.get_bbox(),
                                 inches_per_unit=self.dwgparams.get('inches_per_unit'),
                                 margin=self.dwgparams.get('margin'),
                                 showbbox=self.dwgparams.get('dwgbbox', False))
        if 'bgcolor' in self.dwgparams:
            self.fig.bgcolor(self.dwgparams['bgcolor'])
        self._drawelements()

    def draw(self, show: bool = True,
             canvas=None, backend: Optional[Backends] = None):
        ''' Draw the schematic

            Args:
                show: Show the schematic in a GUI popup window (when
                    outside of a Jupyter inline environment)
                canvas: 'matplotlib', 'svg', or Axis instance to draw on
                backend (deprecated): 'matplotlib' or 'svg'

            Returns:
                schemdraw Figure object
        '''
        if backend:
            warnings.warn('Use of `backend` is deprecated. Use `canvas`.',
                          DeprecationWarning, stacklevel=2)
            canvas = backend

        drawing_stack.push_element(None)

        if canvas is None:
            canvas = self.canvas
        if canvas is None:
            canvas = default_canvas.default_canvas

        if canvas == 'matplotlib':
            self._drawmpl()
        elif hasattr(canvas, 'plot'):
            self._drawmpl(ax=canvas)
        elif canvas == 'svg':
            self._drawsvg()
        else:
            self._drawsvg(canvas)

        if show:
            # Show figure in window if not inline/Jupyter mode
            self.fig.show()  # type: ignore

        if self.outfile is not None:
            self.save(self.outfile)

        return self.fig  # Return Figure and let _repr_ display it

    def save(self, fname: str, transparent: bool = True, dpi: float = 72) -> None:
        ''' Save figure to a file

            Args:
                fname: Filename to save. In Matplotlib backend, the file
                    type is automatically determined from extension
                    (png, svg, jpg). SVG backend only supports saving SVG format.
                transparent: Save as transparent background, if available
                dpi: Dots-per-inch for raster formats
        '''
        if self.fig is None:
            self.draw(show=False)
        self.fig.save(fname, transparent=transparent, dpi=dpi)  # type: ignore

    def get_imagedata(self, fmt: ImageFormat | ImageType = 'svg') -> bytes:
        ''' Get image data as bytes array

            Args:
                fmt: Format or file extension of the image type. SVG backend
                    only supports 'svg' format.

            Returns:
                Image data as bytes
        '''
        if self.canvas == 'svg' and fmt.lower() != 'svg':
            raise ValueError('Format not available in SVG backend.')
        if self.fig is None:
            self.draw(show=False)
        return self.fig.getimage(ext=fmt)  # type: ignore



''' Two-terminal element definitions '''

from __future__ import annotations
from typing import Optional
import math
from typing import Sequence

from .elements import Element, Element2Term, gap
from ..util import Point, linspace
from ..segments import Segment, SegmentArc, SegmentText, SegmentCircle, SegmentPoly


resheight = 0.25      # Resistor height
reswidth = 1.0 / 6   # Full (inner) length of resistor is 1.0 data unit


class ResistorIEEE(Element2Term):
    ''' Resistor (IEEE/U.S. style) '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment(
            [(0, 0), (0.5*reswidth, resheight), (1.5*reswidth, -resheight),
             (2.5*reswidth, resheight), (3.5*reswidth, -resheight),
             (4.5*reswidth, resheight), (5.5*reswidth, -resheight), (6*reswidth, 0)]))


class ResistorIEC(Element2Term):
    ''' Resistor as box (IEC/European style) '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment(
            [(0, 0), (0, resheight), (reswidth*6, resheight),
             (reswidth*6, -resheight), (0, -resheight), (0, 0),
             gap, (reswidth*6, 0)]))


class ResistorVarIEEE(ResistorIEEE):
    ''' Variable resistor (U.S. style) '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment([(1.5*reswidth, -resheight*2), (4.5*reswidth, reswidth*3.5)],
                                     arrow='->', arrowwidth=.16, arrowlength=.2))


class ResistorVarIEC(ResistorIEC):
    ''' Variable resistor (European style) '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment([(1*reswidth, -resheight*2), (5*reswidth, reswidth*3.5)],
                                     arrow='->', arrowwidth=.16, arrowlength=.2))


class Thermistor(ResistorIEC):
    ''' Thermistor '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment([(0, -resheight-.2), (.2, -resheight-.2), (1, resheight+.2)]))


class PhotoresistorIEEE(ResistorIEEE):
    ''' Photo-resistor (U.S. style) '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment([(.7, .75), (.4, .4)], arrow='->',
                                     arrowwidth=.16, arrowlength=.2))
        self.segments.append(Segment([(1, .75), (.7, .4)], arrow='->',
                                     arrowwidth=.16, arrowlength=.2))


class PhotoresistorIEC(ResistorIEC):
    ''' Photo-resistor (European style) '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment([(.7, .75), (.4, .4)], arrow='->',
                                     arrowwidth=.16, arrowlength=.2))
        self.segments.append(Segment([(1, .75), (.7, .4)], arrow='->',
                                     arrowwidth=.16, arrowlength=.2))


class Rshunt(ResistorIEC):
    ''' Shunt Resistor '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment([(0, 0),
                                      (-resheight*1.5, -resheight),
                                      (-resheight*1.5, -resheight*2)]))
        self.segments.append(Segment([(reswidth*6, 0),
                                      (reswidth*6+resheight*1.5, -resheight),
                                      (reswidth*6+resheight*1.5, -resheight*2)]))
        self.anchors['v1'] = (-resheight*1.5, -resheight*2)
        self.anchors['v2'] = (reswidth*6+resheight*1.5, -resheight*2)


class Capacitor(Element2Term):
    ''' Capacitor (flat plates)

        Args:
            polar: Add polarity + sign
    '''
    _element_defaults = {
        'polar': False
    }
    def __init__(self, *, polar: Optional[bool] = None, **kwargs):
        super().__init__(**kwargs)
        capgap = 0.18
        self.segments.append(Segment([(0, 0), gap, (0, resheight), (0, -resheight), gap,
                                     (capgap, resheight), (capgap, -resheight), gap,
                                     (capgap, 0)]))
        if self.params['polar']:
            self.segments.append(SegmentText((-capgap*1.2, capgap), '+'))


class Capacitor2(Element2Term):
    ''' Capacitor (curved bottom plate)

        Args:
            polar: Add polarity + sign
    '''
    _element_defaults = {
        'polar': False
    }
    def __init__(self, *, polar: Optional[bool] = None, **kwargs):
        super().__init__(**kwargs)
        capgap = 0.18
        self.segments.append(Segment([(0, 0), gap, (0, resheight),
                                     (0, -resheight), gap, (capgap, 0)]))
        self.segments.append(SegmentArc((capgap*1.5, 0), width=capgap*1.5,
                                        height=resheight*2.5, theta1=105, theta2=-105))

        if self.params['polar']:
            self.segments.append(SegmentText((-capgap*1.2, capgap), '+'))


class CapacitorVar(Capacitor):
    ''' Variable capacitor '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment([(-2*reswidth, -resheight), (3*reswidth, reswidth*2)],
                                     arrow='->', arrowwidth=.2, arrowlength=.2))


class CapacitorTrim(Capacitor):
    ''' Trim capacitor '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        capgap = 0.18
        # Line endpoints
        p1 = Point((-1.8*reswidth, -resheight))
        p2 = Point((1.8*reswidth+capgap, resheight))
        dx = p2.x - p1.x
        dy = p2.y - p1.y

        self.segments.append(Segment([p1, p2]))

        tlen = .14
        theta = math.atan2(dy, dx) + math.radians(90)
        t1 = Point((p2.x + tlen * math.cos(theta),
                    p2.y + tlen * math.sin(theta)))
        t2 = Point((p2.x + tlen * math.cos(theta-math.pi),
                    p2.y + tlen * math.sin(theta-math.pi)))
        self.segments.append(Segment([t1, t2]))


class Crystal(Element2Term):
    ''' Crystal oscillator '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        xgap = 0.2
        self.segments.append(Segment(
            [(0, 0), gap, (0, resheight), (0, -resheight), gap,
             (xgap/2, resheight), (xgap/2, -resheight), (xgap*1.5, -resheight),
             (xgap*1.5, resheight), (xgap/2, resheight), gap,
             (xgap*2, resheight), (xgap*2, -resheight), gap, (xgap*2, 0)]))


class Diode(Element2Term):
    ''' Diode '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment([(0, 0), gap, (resheight*1.4, resheight),
                                      (resheight*1.4, -resheight), gap, (resheight*1.4, 0)]))
        self.segments.append(SegmentPoly([(0, resheight), (resheight*1.4, 0), (0, -resheight)]))


class Schottky(Diode):
    ''' Schottky Diode '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        schottky_width = 0.1
        self.segments.append(Segment(
            [(resheight*1.4, resheight),
             (resheight*1.4-schottky_width, resheight),
             (resheight*1.4-schottky_width, resheight-schottky_width)]))
        self.segments.append(Segment(
            [(resheight*1.4, -resheight),
             (resheight*1.4+schottky_width, -resheight),
             (resheight*1.4+schottky_width, -resheight+schottky_width)]))


class DiodeTunnel(Diode):
    ''' Tunnel Diode '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        tunnel_width = 0.1
        self.segments.append(Segment([(resheight*1.4, resheight),
                                      (resheight*1.4-tunnel_width, resheight)]))
        self.segments.append(Segment([(resheight*1.4, -resheight),
                                      (resheight*1.4-tunnel_width, -resheight)]))


class DiodeShockley(Element2Term):
    ''' Shockley Diode '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment([(0, 0), (resheight*1.4, 0), (resheight*1.4, resheight),
                                      (resheight*1.4, -resheight), gap, (resheight*1.4, 0)]))
        self.segments.append(Segment([(0, -resheight), (0, resheight), (resheight*1.4, 0)]))


class Zener(Diode):
    ''' Zener Diode '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        zener_width = 0.1
        self.segments.append(Segment([(resheight*1.4, resheight),
                                      (resheight*1.4+zener_width, resheight+zener_width)]))
        self.segments.append(Segment([(resheight*1.4, -resheight),
                                      (resheight*1.4-zener_width, -resheight-zener_width)]))


class Varactor(Element2Term):
    ''' Varactor Diode/Varicap/Variable Capacitance Diode '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        capgap = .13
        self.segments.append(Segment([(0, 0), gap,
                                     (resheight*1.4, resheight), (resheight*1.4, -resheight),
                                     gap,
                                     (resheight*1.4+capgap, resheight), (resheight*1.4+capgap, -resheight),
                                     gap,
                                     (resheight*1.4+capgap, 0)]))
        self.segments.append(SegmentPoly([(0, resheight), (resheight*1.4, 0), (0, -resheight)]))


class LED(Diode):
    ''' Light emitting diode '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment([(resheight, resheight*1.5), (resheight*2, resheight*3.25)],
                                     arrow='->', arrowwidth=.16, arrowlength=.2))
        self.segments.append(Segment([(resheight*.1, resheight*1.5), (resheight*1.1, resheight*3.25)],
                                     arrow='->', arrowwidth=.16, arrowlength=.2))
        self.params['lblloc'] = 'bot'


class LED2(Diode):
    ''' Light emitting diode (curvy light lines) '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        x = linspace(-1, 1)
        y = [-x0*(x0-.7)*(x0+.7)/2 + resheight*2.5 for x0 in x]
        x = linspace(resheight*.75, resheight*1.25)

        theta = 20
        p = [Point((x0, y0)).rotate(theta) for x0, y0 in zip(x, y)]
        p2 = [Point((x0-.2, y0)).rotate(theta) for x0, y0 in zip(x, y)]

        pa = Point((x[0], y[0]+.1)).rotate(theta)
        pa2 = Point((x[0]-.2, y[0]+.1)).rotate(theta)

        self.segments.append(Segment(p))
        self.segments.append(Segment(p2))
        self.segments.append(Segment((p[1], pa), arrow='->', arrowwidth=.15, arrowlength=.2))
        self.segments.append(Segment((p2[1], pa2), arrow='->', arrowwidth=.15, arrowlength=.2))
        self.params['lblloc'] = 'bot'


class Photodiode(Diode):
    ''' Photo-sensitive diode '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        x = linspace(-1, 1)
        y = [-x0*(x0-.7)*(x0+.7)/2 + resheight*2.5 for x0 in x]
        x = linspace(resheight*.75, resheight*1.25)
        theta = 20
        p = [Point((x0, y0)).rotate(theta) for x0, y0 in zip(x, y)]
        p2 = [Point((x0-.2, y0)).rotate(theta) for x0, y0 in zip(x, y)]

        pa = Point((x[-1], y[-1]-.1)).rotate(theta)
        pa2 = Point((x[-1]-.2, y[-1]-.1)).rotate(theta)
        self.segments.append(Segment(p))
        self.segments.append(Segment(p2))
        self.segments.append(Segment((p[-2], pa), arrow='->', arrowwidth=.15, arrowlength=.2))
        self.segments.append(Segment((p2[-2], pa2), arrow='->', arrowwidth=.15, arrowlength=.2))
        self.params['lblloc'] = 'bot'


class PotentiometerIEEE(ResistorIEEE):
    ''' Potentiometer (U.S. style)

        Anchors:
            tap
    '''
    # Ok, this has three terminals, but is works like a two-term with lead extension
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        potheight = .72
        self.anchors['tap'] = (reswidth*3, potheight)
        self.params['lblloc'] = 'bot'
        self.segments.append(Segment([(reswidth*3, potheight), (reswidth*3, reswidth*1.5)],
                                     arrow='->', arrowwidth=.15, arrowlength=.25))


class PotentiometerIEC(ResistorIEC):
    ''' Potentiometer (European style)

        Anchors:
            tap
    '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        potheight = .72
        self.anchors['tap'] = (reswidth*3, potheight)
        self.elmparams['lblloc'] = 'bot'
        self.segments.append(Segment([(reswidth*3, potheight), (reswidth*3, reswidth*2)],
                                     arrow='->', arrowwidth=.15, arrowlength=.22))


class Diac(Element2Term):
    ''' Diac (diode for alternating current) '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment(
            [(0, 0), gap, (resheight*1.4, resheight*1.8),
             (resheight*1.4, -resheight*1.8), gap,
             (0, resheight*1.8), (0, -resheight*1.8), gap, (resheight*1.4, 0)]))
        self.segments.append(SegmentPoly([(0, -resheight-.25), (resheight*1.4, -.25),
                                          (0, -resheight+.25)]))
        self.segments.append(SegmentPoly([(resheight*1.4, resheight+.25), (0, .25),
                                          (resheight*1.4, resheight-.25)]))


class Triac(Diac):
    ''' Triac

        Anchors:
            gate
    '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment([(resheight*1.4, .25), (resheight*1.4+.5, .5)]))
        self.anchors['gate'] = (resheight*1.4+.5, .5)


class SCR(Diode):
    ''' Silicon controlled rectifier (or thyristor)

        Anchors:
            gate
    '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment(
            [(resheight*1.4, 0), (resheight*1.4+.3, -.3), (resheight*1.4+.3, -.5)]))
        self.anchors['gate'] = (resheight*1.4+.3, -.5)


class Memristor(Element2Term):
    ''' Memristor (resistor with memory) '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        mr = 0.2
        self.segments.append(Segment(
            [(0, 0), (mr, 0), (mr, -mr*.75), (mr*2, -mr*.75), (mr*2, mr*.75),
             (mr*3, mr*.75), (mr*3, -mr*.75), (mr*4, -mr*.75), (mr*4, 0),
             (mr*5, 0)]))
        self.segments.append(Segment(
            [(0, mr*1.25), (mr*5, mr*1.25), (mr*5, mr*-1.25), (0, mr*-1.25),
             (0, mr*1.25)]))
        self.segments.append(SegmentPoly(
            [(0, mr*1.25), (0, -mr*1.25), (mr/2, -mr*1.25), (mr/2, mr*1.25)],
            fill='black'))


class Memristor2(Element2Term):
    ''' Memristor (resistor with memory), alternate style '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        mr = 0.2
        mrv = .25
        self.segments.append(Segment(
            [(0, 0), (0, mrv), (mr, mrv), (mr, -mrv), (mr*2, -mrv), (mr*2, mrv),
             (mr*3, mrv), (mr*3, -mrv), (mr*4, -mrv), (mr*4, mrv),
             (mr*5, mrv), (mr*5, -mrv), (mr*6, -mrv), (mr*6, 0),
             (mr*7, 0)]))


class Josephson(Element2Term):
    ''' Josephson Junction '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment(
            [(0, 0), gap, (-resheight, resheight), (resheight, -resheight),
             gap, (resheight, resheight), (-resheight, -resheight),
             gap, (0, 0)]))


class FuseUS(Element2Term):
    ''' Fuse (U.S. Style)

        Args:
            dots: Show dots on connections to fuse
    '''
    _element_defaults = {
        'dots': True
    }
    def __init__(self, *, dots: Optional[bool] = None, **kwargs):
        super().__init__(**kwargs)
        fuser = .12
        fusex = linspace(fuser*2, 1+fuser)
        fusey = [math.sin(x) * resheight for x in linspace(0, 2*math.pi)]
        self.segments.append(Segment(list(zip(fusex, fusey))))
        self.segments.append(Segment([(0, 0), gap, (1+fuser*3, 0)]))
        if self.params['dots']:
            self.fill(kwargs.get('fill', 'bg'))

    def fill(self, color: bool | str = True) -> 'Element':
        ''' Set element fill '''
        fuser = .12
        self.segments.append(SegmentCircle(
            (fuser, 0), fuser, zorder=4, fill=color))
        self.segments.append(SegmentCircle(
            (fuser*2+1, 0), fuser, zorder=4, fill=color))
        super().fill(color)
        return self


class FuseIEEE(ResistorIEC):
    ''' Fuse (IEEE Style) '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment([(0, 0), (reswidth*6, 0)]))


class FuseIEC(ResistorIEC):
    ''' Fuse (IEC Style) '''    
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        dx = resheight*.66
        self.segments.append(Segment([(dx, resheight), (dx, -resheight)]))
        self.segments.append(Segment([(6*reswidth-dx, resheight),
                                      (6*reswidth-dx, -resheight)]))


class Breaker(Element2Term):
    ''' Circuit breaker

        Args:
            dots: Show connection dots
    '''
    _element_defaults = {
        'dots': True
    }
    def __init__(self, *, dots: Optional[bool] = None, **kwargs):
        super().__init__(**kwargs)
        theta1 = 25 if dots else 10
        theta2 = 155 if dots else 170
        self.segments.append(Segment(
            [(0, 0), gap, (1, 0)]))
        self.segments.append(SegmentArc(
            (.5, 0), 1, .65, theta1=theta1, theta2=theta2))
        if self.params['dots']:
            rad = .12
            self.segments.append(SegmentCircle((rad, 0), rad, zorder=4))
            self.segments.append(SegmentCircle((1-rad, 0), rad, zorder=4))


def cycloid(loops: int = 4, ofst: Sequence[float] = (0, 0),
            a: float = .06, b: float = .19, norm: bool = True,
            vertical: bool = False,
            flip: bool = False) -> list[tuple[float, float]]:
    ''' Generate a prolate cycloid (inductor spiral) that
        will always start and end at y=0.

        Args:
            loops: Number of loops
            a, b: Parameters. b>a for prolate (loopy) cycloid
            norm: Normalize the length to 1
            vertical: draw vertically (swap x, y)
            flip: flip orientation (invert y)

        Returns:
            List of (x, y) coordinates defining the cycloid
    '''
    yint = math.acos(a/b)  # y-intercept
    t = linspace(yint, 2*(loops+1)*math.pi-yint, num=loops*50)
    x = [a*t0 - b*math.sin(t0) for t0 in t]
    y = [a - b*math.cos(t0) for t0 in t]
    x = [xx - x[0] for xx in x]  # Shift to start at 0,0

    if norm:
        x = [xx / (x[-1]-x[0]) for xx in x]  # Normalize length to 1

    if flip:
        y = [-yy for yy in y]

    y = [yy * (max(y)-min(y))/(resheight) for yy in y]  # Normalize to resistor width

    if vertical:
        x, y = y, x

    x = [xx + ofst[0] for xx in x]
    y = [yy + ofst[1] for yy in y]

    path = list(zip(x, y))
    return path


class Inductor(Element2Term):
    ''' Inductor '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        ind_w = .25
        self.segments.append(Segment([(0, 0), gap, (1, 0)]))
        for i in range(4):
            self.segments.append(SegmentArc(
                [(i*2+1)*ind_w/2, 0], theta1=0, theta2=180,
                width=ind_w, height=ind_w))


class Inductor2(Element2Term):
    ''' Inductor, drawn as cycloid (loopy)

        Args:
            loops: Number of inductor loops [default: 4]
    '''
    _element_defaults = {
        'loops': 4,
    }
    def __init__(self, *, loops: Optional[int] = None, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment(cycloid(loops=self.params['loops'])))


class CPE(Element2Term):
    ''' Constant Phase Element '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        capgap = 0.25
        offset = 0.25
        self.segments.append(Segment([(0, 0), (-offset, 0), gap, (offset, 0),
                                      (0, -resheight), gap, (capgap, 0)]))
        self.segments.append(Segment([(0, 0), (-offset, 0), gap, (offset, 0),
                                      (0, resheight), gap, (capgap, 0)]))
        self.segments.append(Segment([(0, 0), (-offset, resheight), gap,
                                      (0, -resheight), gap, (capgap, 0)]))
        self.segments.append(Segment([(0, 0), (-offset, -resheight), gap,
                                      (0, -resheight), gap, (capgap, 0)]))


class SparkGap(Element2Term):
    ''' Spark Gap '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment([(0, 0), (.3, 0), gap, (.7, 0), (1, 0)]))
        # Arrow coords overlap a bit since default arrow is offset by linewidth
        self.segments.append(Segment([(.3, 0), (.52, 0)], arrow='->', arrowwidth=.2))
        self.segments.append(Segment([(.7, 0), (.48, 0)], arrow='->', arrowwidth=.2))


class Nullator(Element2Term):
    ''' Nullator

        This element does not support filling
    '''
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.segments.append(Segment([[0, 0], [0, 0], gap, [1, 0], [1, 0]]))
        self.segments.append(SegmentArc(center=[0.5,0], width=1, height=0.5, theta1=0, theta2=360))
        self.elmparams['theta'] = 90


class Norator(Element2Term):
    ''' Norator '''
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.segments.append(Segment([[0, 0], [0, 0], gap, [1, 0], [1, 0]]))
        self.segments.append(SegmentCircle([0.25, 0], 0.25,))
        self.segments.append(SegmentCircle([1-0.25, 0], 0.25,))
        self.elmparams['theta'] = 90


class CurrentMirror(Element2Term):
    ''' Current mirror with optional common terminal

        Anchors:
            * scommon
    '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment([(0, 0), (0, 0), gap, (1, 0), (1, 0)]))
        # in order to prevent graphical glitches due to overlapping outlines when filling, draw fill before outline
        # fill is drawn here. The outline is disabled to prevent anti-aliasing glitches
        # ideally, the outline would be disabled with ls='None', but this does not work in testing, so lw=1e-9 is here
        # as a fallback. 0 cannot be used because it would be ignored by the style logic
        self.segments.append(SegmentCircle([0.3, 0], 0.3, lw=1e-9, ls="None"))
        self.segments.append(SegmentCircle([1 - 0.3, 0], 0.3, lw=1e-9, ls="None"))
        # outline is drawn here, without fill to prevent the glitch
        self.segments.append(SegmentCircle([0.3, 0], 0.3, fill=False))
        self.segments.append(SegmentCircle([1 - 0.3, 0], 0.3, fill=False))
        self.elmparams['theta'] = 90
        self.anchors['common'] = (0.5, math.sqrt(0.3**2 - 0.2**2))


class VoltageMirror(Element2Term):
    ''' Voltage mirror with optional common terminal

        This element does not support filling

        Anchors:
            * scommon
    '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment([[0, 0], [0, 0], gap, [1, 0], [1, 0]]))
        self.segments.append(SegmentArc(center=[0.7, 0], width=0.6, height=0.3, theta1=0, theta2=360))
        self.segments.append(SegmentArc(center=[0.3, 0], width=0.6, height=0.3, theta1=0, theta2=360))
        self.elmparams['theta'] = 90
        self.anchors['common'] = (0.5, math.sqrt(1 - 0.2**2/0.3**2)*0.15)


# default to IEEE style
Resistor = ResistorIEEE
ResistorVar = ResistorVarIEEE
Photoresistor = PhotoresistorIEEE
Potentiometer = PotentiometerIEEE
Fuse = FuseIEEE

# Old names
RBox = ResistorIEC
RBoxVar = ResistorVarIEC
PotBox = PotentiometerIEC
PhotoresistorBox = PhotoresistorIEC



''' Integrated Circuit Element '''

from __future__ import annotations
from typing import Optional, Sequence, Tuple, cast
from collections import namedtuple
import math
from dataclasses import dataclass, replace

from ..segments import Segment, SegmentText, SegmentCircle, SegmentPoly, SegmentType
from ..elements import Element
from ..util import linspace, Point
from ..types import XY, Side, Halign, Valign
from ..backends.svg import text_size


@dataclass
class IcPin:
    ''' Integrated Circuit Pin

        Args:
            name: Input/output name (inside the box)
            pin: Pin name/number (outside the box)
            side: Side of box for the pin, 'left', 'right', 'top', 'bottom'
            pos: Pin position along the side, fraction from 0-1
            slot: Slot definition of pin location, given in 'X/Y' format.
                '2/4' is the second pin on a side with 4 pins.
            invert:Add an invert bubble to the pin
            invertradius: Radius of invert bubble
            color: Color for the pin and label
            rotation: Rotation for label text
            anchorname: Named anchor for the pin
     '''
    name: str | None = None
    pin: str | None = None
    side: Side = 'L'
    pos: float | None = None
    slot: str | None = None
    invert: bool = False
    invertradius: float = 0.15
    color: str | None = None
    rotation: float = 0
    anchorname: str | None = None
    lblsize: float | None = None
    pinlblsize: float | None = None


@dataclass
class IcSide:
    ''' Pin layout parameters for one side of an Ic '''
    spacing: float = 0.0  # 0 for auto
    pad: float = .25
    leadlen: float = 0.5
    label_ofst: float = 0.15
    label_size: float = 14.
    pinlabel_ofst: float = 0.05
    pinlabel_size: float = 11.


IcBox = namedtuple('IcBox', 'w h y1 y2')


class Ic(Element):
    _element_defaults = {
        'pinspacing': 0.0,
        'edgepadH': 0.5,
        'edgepadW': 0.5,
        'leadlen': 0.5,
        'lsize': 14,
        'lofst': 0.15,
        'plblofst': 0.05,
        'plblsize': 11
    }
    def __init__(self,
                 size: Optional[XY] = None,
                 pins: Optional[Sequence[IcPin]] = None,
                 slant: float = 0,
                 **kwargs):
        super().__init__(**kwargs)
        self.size = size if size is not None else self.params.get('size')
        self._sizeauto: Optional[tuple[float, float]] = None
        self.slant = slant
        self.pins: dict[Side, list[IcPin]] = {'L': [], 'R': [], 'T': [], 'B': []}
        self.usersides: dict[Side, IcSide] = {}
        self.sides: dict[Side, IcSide] = {}
        self._dflt_side = IcSide(
            self.params.get('pinspacing', 0),
            self.params.get('edgepadH', .25),
            self.params.get('leadlen', .5),
            self.params.get('lofst', .15),
            self.params.get('lsize', 14),
            self.params.get('plblofst', .05),
            self.params.get('plblsize', 11))

        if pins is not None:
            for pin in pins:
                side = cast(Side, pin.side[0].upper())
                self.pins[side].append(pin)

        self._icbox = IcBox(0,0,0,0)
        self._setsize()

    def pin(self,
            side: Side = 'L',
            name: str | None = None,
            pin: str | None = None,
            pos: float | None = None,
            slot: str | None = None,
            invert: bool = False,
            invertradius: float = 0.15,
            color: str | None = None,
            rotation: float = 0,
            anchorname: str | None = None,
            lblsize: float | None = None):
        ''' Add a pin to the IC

            Args:
                name: Input/output name (inside the box)
                pin: Pin name/number (outside the box)
                side: Side of box for the pin, 'left', 'right', 'top', 'bottom'
                pos: Pin position along the side, fraction from 0-1
                slot: Slot definition of pin location, given in 'X/Y' format.
                    '2/4' is the second pin on a side with 4 pins.
                invert:Add an invert bubble to the pin
                invertradius: Radius of invert bubble
                color: Color for the pin and label
                rotation: Rotation for label text
                anchorname: Named anchor for the pin
                lblsize: Font size for label
        '''
        side = cast(Side, side[0].upper())
        self.pins[side].append(IcPin(name, pin, side, pos, slot, invert,
                                    invertradius, color, rotation, anchorname, lblsize))
        self._setsize()
        return self

    def side(self,
             side: Side = 'L',
             spacing: float = 0,
             pad: float = 0.5,
             leadlen: float = .5,
             label_ofst: float = 0.15,
             label_size: float = 14.,
             pinlabel_ofst: float = 0.05,
             pinlabel_size: float = 11.):
        ''' Set parameters for spacing/layout of one side

            Args:
                side: Side of box to define
                spacing: Distance between pins
                pad: Distance from box edge to first pin
                leadlen: Length of pin lead extensions
                label_ofst: Offset between box and label (inside box)
                label_size: Font size of label (inside box)
                pinlabel_ofst: Offset between box and pin label (outside box)
                pinlabel_size: Font size of pin label (outside box)
        '''
        side = cast(Side, side[0].upper())
        self.usersides[side] = IcSide(spacing, pad, leadlen,
                                      label_ofst, label_size,
                                      pinlabel_ofst, pinlabel_size)
        self._setsize()
        return self

    def _countpins(self) -> dict[Side, int]:
        ''' Count the number of pins (or slots) on each side '''
        pincount = {}
        for side, pins in self.pins.items():
            slotnames = [p.slot for p in pins]
            # Add a 0 - can't max an empty list
            slots = [int(p.split('/')[1]) for p in slotnames if p is not None] + [0]
            pincount[side] = max(len(pins), max(slots))
        return pincount

    def _autosize(self) -> None:
        ''' Determine size of box if not provided '''
        lengths: dict[str, float] = {}
        labelwidths: dict[str, float] = {}

        # Determine length of each side given spacing and pad
        for side in ['L', 'R', 'T', 'B']:
            side = cast(Side, side)
            sideparam = replace(self.usersides.get(side, self._dflt_side))
            self.sides[side] = sideparam
            if sideparam.spacing == 0:
                sideparam.spacing = 0.6
            lengths[side] = sideparam.pad * 2 + sideparam.spacing * (self.pincount[side] - 1)

            # Expand to fit pin labels if necessary
            labelw = 0.
            for p in self.pins[side]:
                if p.name:
                    lblsize = p.lblsize if p.lblsize else self.params['lsize']
                    labelw = max(labelw, text_size(p.name, size=lblsize)[0]/72*2)
            labelwidths[side] = labelw

        # Box is larger of the two sides, with minimum of 2
        labelw = labelwidths['L'] + labelwidths['R'] + 4*self.params['lofst']
        boxh = max(lengths.get('L', 0.), lengths.get('R', 0.), 2+self.params['edgepadH'])
        boxw = max(lengths.get('T', 0.), lengths.get('B', 0.), 2+self.params['edgepadW'], labelw)
        self._sizeauto = boxw, boxh

        # Adjust pad for the shorter of the two parallel sides
        for side in ['L', 'R']:
            side = cast(Side, side)
            sideparam = self.sides.get(side, self._dflt_side)
            sideparam.pad = (boxh - sideparam.spacing * (self.pincount[side] - 1)) / 2

        for side in ['T', 'B']:
            side = cast(Side, side)
            sideparam = self.sides.get(side, self._dflt_side)
            sideparam.pad = (boxw - sideparam.spacing * (self.pincount[side] - 1)) / 2

    def _autopinlayout(self) -> None:
        ''' Determine pin layout when box size is specified '''
        for side in ['L', 'R', 'T', 'B']:
            side = cast(Side, side)
            sideparam = replace(self.usersides.get(side, self._dflt_side))
            self.sides[side] = sideparam
            length = self.size[0] if side in ['T', 'B'] else self.size[1]
            pad = sideparam.pad
            if sideparam.spacing == 0:
                #use PAD to evenly space pins over (length-2*pad)
                if self.pincount[side] > 1:
                    sideparam.spacing = (length - 2*pad) / (self.pincount[side] - 1)
                else:
                    sideparam.pad = length/2
            else:
                # Addjust PAD center the group of pins
                sideparam.pad = (length - sideparam.spacing*(self.pincount[side]-1)) / 2

    def _drawbox(self) -> IcBox:
        ''' Draw main box and return its size '''
        if self.size:
            w, h = self.size
        elif self._sizeauto:
            w, h = self._sizeauto
        else:
            w, h = (2, 3)

        tanslant = w * math.tan(math.radians(self.slant))
        if self.slant > 0:
            y1 = 0 - tanslant
            y2 = h + tanslant
            path = [Point((0, 0)), Point((w, y1)), Point((w, y2)), Point((0, h)), Point((0, 0))]
        elif self.slant < 0:
            y1 = 0 + tanslant
            y2 = h - tanslant
            path = [Point((0, y1)), Point((w, 0)), Point((w, h)), Point((0, y2)), Point((0, y1))]
        else:
            y1 = 0
            y2 = h
            path = [Point((0, 0)), Point((w, 0)), Point((w, h)), Point((0, h)), Point((0, 0))]
        self.segments.append(Segment(path))
        return IcBox(w, h, y1, y2)

    def _pinpos(self, side: Side, pin: IcPin, num: int) -> Point:
        ''' Get XY position of pin '''
        sidesetup = self.sides.get(side, self._dflt_side)
        spacing = sidesetup.spacing if sidesetup.spacing > 0 else 0.6
        if pin.slot:
            num = int(pin.slot.split('/')[0]) - 1
        z = sidesetup.pad + num*spacing
        if pin.pos:
            z = sidesetup.pad + pin.pos * (self.pincount[side]-1)*spacing
        xy = {'L': Point((0, z)),
              'R': Point((self._icbox.w, z)),
              'T': Point((z, self._icbox.h)),
              'B': Point((z, 0))
              }.get(side, Point((0, 0)))

        # Adjust pin position for slant
        if side == 'T' and self.slant > 0:
            xy = Point((xy[0], xy[1] - xy[0] * math.tan(-math.radians(self.slant))))
        elif side == 'T' and self.slant < 0:
            xy = Point(((xy[0], xy[1] + (self._icbox.y2-self._icbox.h) - xy[0] * math.tan(-math.radians(self.slant)))))
        elif side == 'B' and self.slant < 0:
            xy = Point((xy[0], xy[1] - (self._icbox.y2-self._icbox.h) - xy[0] * math.tan(math.radians(self.slant))))
        elif side == 'B' and self.slant > 0:
            xy = Point((xy[0], xy[1] - xy[0] * math.tan(math.radians(self.slant))))
        return xy

    def _drawpin(self, side: Side, pin: IcPin, num: int) -> None:
        ''' Draw one pin and its labels '''
        # num is index of where pin is along side
        sidesetup = self.sides.get(side, self._dflt_side)
        xy = self._pinpos(side, pin, num)
        leadext = {'L': Point((-sidesetup.leadlen, 0)),
                   'R': Point((sidesetup.leadlen, 0)),
                   'T': Point((0, sidesetup.leadlen)),
                   'B': Point((0, -sidesetup.leadlen))}.get(side, Point((0, 0)))

        # Anchor
        anchorpos = xy+leadext
        self.anchors[f'in{side[0].upper()}{num+1}'] = anchorpos
        if pin.anchorname:
            self.anchors[pin.anchorname] = anchorpos
        elif pin.name:
            if pin.name == '>':
                self.anchors['CLK'] = anchorpos
            self.anchors[pin.name] = anchorpos
        if pin.pin:
            self.anchors[f'pin{pin.pin}'] = anchorpos

        # Lead        
        if sidesetup.leadlen > 0:
            if pin.invert:  # Add invert-bubble
                invertradius = pin.invertradius
                invertofst = {'L': Point((-invertradius, 0)),
                              'R': Point((invertradius, 0)),
                              'T': Point((0, invertradius)),
                              'B': Point((0, -invertradius))}.get(side, Point((0, 0)))
                self.segments.append(SegmentCircle(
                    xy+invertofst, invertradius))
                self.segments.append(Segment([xy+invertofst*2, xy+leadext]))
            else:
                self.segments.append(Segment([xy, xy+leadext]))

        # Pin Number (outside the box)
        if pin.pin and pin.pin != '':
            # Account for any invert-bubbles
            invertradius = pin.invertradius * pin.invert
            plbl = sidesetup.pinlabel_ofst
            pofst = {'L': Point((-plbl-invertradius*2, plbl)),
                     'R': Point((plbl+invertradius*2, plbl)),
                     'T': Point((plbl, plbl+invertradius*2)),
                     'B': Point((plbl, -plbl-invertradius*2))
                     }.get(side)

            align = cast(Optional[Tuple[Halign, Valign]],
                         {'L': ('right', 'bottom'),
                          'R': ('left', 'bottom'),
                          'T': ('left', 'bottom'),
                          'B': ('left', 'top')}.get(side))
            self.segments.append(SegmentText(
                pos=xy+pofst,
                label=pin.pin,
                align=align,
                fontsize=pin.pinlblsize if pin.pinlblsize is not None else sidesetup.pinlabel_size))

        if pin.name == '>':
            self._drawclkpin(xy, leadext, side, pin, num)
            return

        # Label (inside the box)
        if pin.name and pin.name != '':
            pofst = {'L': Point((sidesetup.label_ofst, 0)),
                     'R': Point((-sidesetup.label_ofst, 0)),
                     'T': Point((0, -sidesetup.label_ofst)),
                     'B': Point((0, sidesetup.label_ofst))}.get(side)

            align = cast(Optional[Tuple[Halign, Valign]],
                         {'L': ('left', 'center'),
                          'R': ('right', 'center'),
                          'T': ('center', 'top'),
                          'B': ('center', 'bottom')}.get(side))

            self.segments.append(SegmentText(
                pos=xy+pofst,
                label=pin.name,
                align=align,
                fontsize=pin.lblsize if pin.lblsize is not None else sidesetup.label_size,
                color=pin.color,
                rotation=pin.rotation,
                rotation_mode='default'))

    def _drawclkpin(self, xy: Point, leadext: Point,
                    side: Side, pin: IcPin, num: int) -> None:
        ''' Draw clock pin > '''
        sidesetup = self.sides.get(side, self._dflt_side)
        sidesetup.label_size
        clkw, clkh = 0.4 * sidesetup.label_size/16, 0.2 * sidesetup.label_size/16
        if side in ['T', 'B']:
            clkw = math.copysign(clkw, leadext[1]) if leadext[1] != 0 else clkw
            clkpath = [Point((xy[0]-clkh, xy[1])),
                        Point((xy[0], xy[1]-clkw)),
                        Point((xy[0]+clkh, xy[1]))]
        else:
            clkw = math.copysign(clkw, -leadext[0]) if leadext[0] != 0 else clkw
            clkpath = [Point((xy[0], xy[1]+clkh)),
                        Point((xy[0]+clkw, xy[1])),
                        Point((xy[0], xy[1]-clkh))]
        self.segments.append(Segment(clkpath))

    def _drawpins(self) -> None:
        ''' Draw all the pins '''
        for side, pins in self.pins.items():
            for i, pin in enumerate(pins):
                self._drawpin(side, pin, i)

    def _setsize(self) -> None:
        ''' Set size based on pins added so far '''
        self.pincount = self._countpins()
        if self.size is None:
            self._autosize()
        else:
            self._autopinlayout()
        self.pinspacing = {'L': self.sides['L'].spacing,
                           'R': self.sides['R'].spacing,
                           'T': self.sides['T'].spacing,
                           'B': self.sides['B'].spacing}

    def _place(self, dwgxy: XY, dwgtheta: float, **dwgparams) -> tuple[Point, float]:
        self._icbox = self._drawbox()
        self._drawpins()
        self.elmparams['lblloc'] = 'center'
        self.anchors['center'] = (self._icbox.w/2, self._icbox.h/2)
        return super()._place(dwgxy, dwgtheta, **dwgparams)


class Multiplexer(Ic):
    ''' Multiplexer

        Args:
            demux: Draw as demultiplexer
            size: (Width, Height) of box
            pins: List of IcPin instances defining the inputs/outputs
            slant: Slant angle of top/bottom edges

        Keyword Args:            
            pinspacing: Spacing between pins [default: 0.6]
            edgepadH: Padding between top/bottom and first pin [default: 0.25]
            edgepadW: Padding between left/right and first pin [default: 0.25]
            lofst: Offset between edge and label (inside box) [default: 0.15]
            lsize: Font size of labels (inside box) [default: 14]
            plblofst: Offset between edge and pin label (outside box) [default: 0.05]
            plblsize: Font size of pin labels (outside box) [default: 11]

        If a pin is named '>', it will be drawn as a proper clock signal input.

        Anchors:
            * inL[X] - Each pin on left side
            * inR[X] - Each pin on right side
            * inT[X] - Each pin on top side
            * inB[X] - Each pin on bottom side
            * pin[X] - Each pin with a number
            * CLK (if clock pin is defined with '>' name)

        Pins with names are also defined as anchors (if the name does
        not conflict with other attributes).
    '''
    def __init__(self,
                 demux: bool = False,
                 size: Optional[XY] = None,
                 pins: Optional[Sequence[IcPin]] = None,
                 slant: float = 25,
                 **kwargs):
        slant = -slant if not demux else slant
        super().__init__(size=size,
                         pins=pins,
                         slant=slant,
                         **kwargs)


class IcDIP(Element):
    ''' Dual-inline Package Integrated Circuit.

        Args:
            names: List of names for each pin to display inside the box

        Keyword Args:
            pins: number of pins [default: 8]
            notch: Show the notch at top of box [default: True]
            width: Width of the box [default: 3]
            pinw: Width and height of each pin [default: 0.6]
            spacing: Distance between each pin [default: 0.5]
            number: Show pin numbers inside each pin [default: True]
            fontsize: Size for pin name labels [default: 12]
            pfontsize: Size for pin number labels [default: 10]

        Anchors:
            * p[x]  - Each pin
            * p[x]_in  - Inside contact for each pin

        If signal names are provided, they will also be added as anchors
        along with _in inside variants.
    '''
    _element_defaults = {
        'pins': 8,
        'notch': True,
        'width': 3.,
        'pinw': 0.6,
        'spacing': 0.5,
        'number': True,
        'fontsize': 12,
        'pfontsize': 10
    }
    def __init__(self, *,
                 pins: Optional[int] = None,
                 names: Optional[Sequence[str]] = None,
                 notch: Optional[bool] = None,
                 width: Optional[float] = None,
                 pinw: Optional[float] = None,
                 spacing: Optional[float] = None,
                 number: Optional[bool] = None,
                 fontsize: Optional[float]= None,
                 pfontsize: Optional[float] = None,
                 **kwargs):
        super().__init__(**kwargs)
        w: float = self.params['width']
        pw: float = self.params['pinw']
        space: float = self.params['spacing']
        if self.params['pins'] % 2 == 1:
            raise ValueError('pins must be even')

        height = self.params['pins']/2*pw + space*(self.params['pins']/2+1)
        self.segments.append(SegmentPoly(((0, 0), (w, 0), (w, height), (0, height))))

        if self.params['notch']:
            c = w/2
            theta = linspace(-math.pi, 0)
            notchr = pw/2
            notchx = [c + notchr * math.cos(t) for t in theta]
            notchy = [height + notchr * math.sin(t) for t in theta]
            self.segments.append(Segment(list(zip(notchx, notchy))))

        for i in range(self.params['pins']//2):
            y1 = space + space*i+i*pw
            y2 = space + space*i+(i+1)*pw
            ymid = (y1+y2)/2

            # Left
            self.segments.append(SegmentPoly(((0, y1), (0, y2), (-pw, y2), (-pw, y1))))
            pnum = self.params['pins']//2 - i
            self.anchors[f'p{pnum}'] = (-pw, ymid)
            self.anchors[f'p{pnum}_in'] = (0, ymid)
            if self.params['number']:
                self.segments.append(SegmentText((-pw/2, ymid), str(pnum), fontsize=self.params['pfontsize']))
            if names:
                self.segments.append(SegmentText(
                    (.1, ymid), names[pnum-1], align=('left', 'center'), fontsize=self.params['fontsize']))
                self.anchors[names[pnum-1]] = (-pw, ymid)
                self.anchors[f'{names[pnum-1]}_in'] = (0, ymid)

            # Right
            self.segments.append(SegmentPoly(((w, y1), (w, y2), (w+pw, y2), (w+pw, y1))))
            pnum = self.params['pins']//2 + i + 1
            self.anchors[f'p{pnum}'] = (w+pw, ymid)
            self.anchors[f'p{pnum}_in'] = (w, ymid)
            if self.params['number']:
                self.segments.append(SegmentText((w+pw/2, ymid), str(pnum), fontsize=self.params['pfontsize']))
            if names:
                self.segments.append(SegmentText(
                    (w-.1, ymid), names[pnum-1], align=('right', 'center'), fontsize=self.params['fontsize']))
                self.anchors[names[pnum-1]] = (w+pw, ymid)
                self.anchors[f'{names[pnum-1]}_in'] = (w, ymid)


class DFlipFlop(Ic):
    ''' D-Type Flip Flop

        Args:
            preclr: Show preset and clear inputs
            preclrinvert: Add invert bubble to preset and clear inputs

        Keyword Args:
            size: Size of the box [default: (2, 3)]

        Anchors:
            * D
            * CLK
            * Q
            * Qbar
            * PRE
            * CLR
    '''
    _element_defaults = {
        'size': (2, 3)
    }
    def __init__(self, preclr: bool = False, preclrinvert: bool = True, **kwargs):
        pins = [IcPin('D', side='left', slot='2/2'),
                IcPin('>', side='left', slot='1/2'),
                IcPin('Q', side='right', slot='2/2'),
                IcPin(r'$\overline{Q}$', side='right', slot='1/2', anchorname='Qbar')]

        if preclr:
            pins.extend([IcPin('PRE', side='top', invert=preclrinvert),
                         IcPin('CLR', side='bottom', invert=preclrinvert)])
        super().__init__(pins=pins, **kwargs)


class JKFlipFlop(Ic):
    ''' J-K Flip Flop

        Args:
            preclr: Show preset and clear inputs
            preclrinvert: Add invert bubble to preset and clear inputs
            
        Keyword Args:
            size: Size of the box [default: (2, 3)]

        Anchors:
            * J
            * K
            * CLK
            * Q
            * Qbar
            * PRE
            * CLR
    '''
    _element_defaults = {
        'size': (2, 3)
    }
    def __init__(self, preclr: bool = False, preclrinvert: bool = True, **kwargs):
        pins = [IcPin('J', side='left', slot='3/3'),
                IcPin('>', side='left', slot='2/3'),
                IcPin('K', side='left', slot='1/3'),
                IcPin('Q', side='right', slot='3/3'),
                IcPin(r'$\overline{Q}$', side='right', slot='1/3', anchorname='Qbar')]

        if preclr:
            pins.extend([IcPin('PRE', side='top', invert=preclrinvert),
                         IcPin('CLR', side='bottom', invert=preclrinvert)])

        super().__init__(pins=pins, **kwargs)


class VoltageRegulator(Ic):
    ''' Voltage regulator

        Keyword Args:
            size: Size of the box [default: (2, 1.5)]

        Anchors:
            * in
            * out
            * gnd
    '''
    _element_defaults = {
        'size': (2, 1.5)
    }
    def __init__(self, **kwargs):
        pins = [IcPin('in', side='left', slot='3/3'),
                IcPin('out', side='right', slot='3/3'),
                IcPin('gnd', side='bottom')]
        super().__init__(pins=pins, **kwargs)


class Ic555(Ic):
    _element_defaults = {
        'edgepadW': 0.5,
        'edgepadH': 1,
        'pinspacing': 1.5,
        'leadlen': 1
    }
    def __init__(self, **kwargs):
        pins = [IcPin(name='TRG', side='left', pin='2'),
                IcPin(name='THR', side='left', pin='6'),
                IcPin(name='DIS', side='left', pin='7'),
                IcPin(name='CTL', side='right', pin='5'),
                IcPin(name='OUT', side='right', pin='3'),
                IcPin(name='RST', side='top', pin='4'),
                IcPin(name='Vcc', side='top', pin='8'),
                IcPin(name='GND', side='bot', pin='1')]
        super().__init__(pins=pins, label='555')


def sevensegdigit(bottom: float = 0, left: float = 0,
                  seglen: float = 1.5, segw: float = 0.3, spacing: float = 0.12,
                  decimal: bool = False, digit: int | str = 8,
                  segcolor: str = 'red', tilt: float = 10, labelsegments: bool = True) -> list[SegmentType]:
    ''' Generate drawing segments for a 7-segment display digit. Use for
        building new elements incorporating a 7-segment display.

        Args:
            bottom: Location of bottom of digit
            left: Location of left side of digit
            seglen: Length of one segment
            segw: Width of one segment
            spacing: Distance between segments in corners
            decimal: Show decimal point segment
            digit: Number to display
            segcolor: Color of segments
            tilt: Tilt angle in degrees
            labelsegments: Add a-g labels to each segment
            anode: Add common anode pin
            cathode: Add common cathode pin
            size: Total size of the box

        Returns:
            List of Segments making the digit
    '''
    halfw = segw/2  # Half segment width
    halfspace = spacing/2
    tilt = math.radians(tilt)

    # Straight (non-tilted) segments - Horizontal
    segDx = [halfspace, halfspace+halfw, seglen-halfspace-halfw,
             seglen-halfspace, seglen-halfspace-halfw, halfspace+halfw]
    segDy = [0, halfw, halfw, 0, -halfw, -halfw]
    segAx = segDx
    segGx = segDx
    segAy = [k + seglen*2 for k in segDy]
    segGy = [k + seglen for k in segDy]
    # - Vertical
    segEx = [0, halfw, halfw, 0, -halfw, -halfw]
    segEy = [halfspace, halfspace+halfw, seglen-halfspace-halfw,
             seglen-halfspace,  seglen-halfspace-halfw, halfspace+halfw]
    segFx = segEx
    segFy = [k + seglen for k in segEy]
    segBx = [k + seglen for k in segEx]
    segBy = segFy
    segCx = segBx
    segCy = segEy

    # Label positions
    segAlabel = (seglen/2 + halfspace, seglen*2+halfw)
    segGlabel = (seglen/2 + halfspace, seglen+halfw)
    segDlabel = (seglen/2 + halfspace, halfw)
    segBlabel = (seglen + halfw + halfspace, seglen*1.5)
    segClabel = (seglen + halfw + halfspace, seglen*.5)
    segElabel = (-halfw - halfspace, seglen*.5)
    segFlabel = (-halfw - halfspace, seglen*1.5)

    # Apply shear
    if tilt != 0:
        lam = math.sin(tilt)
        segAx = [k + lam*i for k, i in zip(segAx, segAy)]
        segBx = [k + lam*i for k, i in zip(segBx, segBy)]
        segCx = [k + lam*i for k, i in zip(segCx, segCy)]
        segDx = [k + lam*i for k, i in zip(segDx, segDy)]
        segEx = [k + lam*i for k, i in zip(segEx, segEy)]
        segFx = [k + lam*i for k, i in zip(segFx, segFy)]
        segGx = [k + lam*i for k, i in zip(segGx, segGy)]
        def shiftlabel(label):
            return label[0] + label[1]*lam, label[1]
        segAlabel = shiftlabel(segAlabel)
        segGlabel = shiftlabel(segGlabel)
        segDlabel = shiftlabel(segDlabel)
        segBlabel = shiftlabel(segBlabel)
        segClabel = shiftlabel(segClabel)
        segElabel = shiftlabel(segElabel)
        segFlabel = shiftlabel(segFlabel)
        left -= lam*seglen

    # Translate to final position
    segDx = [k + left for k in segDx]
    segDy = [k + bottom for k in segDy]
    segGx = [k + left for k in segGx]
    segGy = [k + bottom for k in segGy]
    segAx = [k + left for k in segAx]
    segAy = [k + bottom for k in segAy]
    segEx = [k + left for k in segEx]
    segEy = [k + bottom for k in segEy]
    segFx = [k + left for k in segFx]
    segFy = [k + bottom for k in segFy]
    segBx = [k + left for k in segBx]
    segBy = [k + bottom for k in segBy]
    segCx = [k + left for k in segCx]
    segCy = [k + bottom for k in segCy]
    segAlabel = (segAlabel[0] + left, segAlabel[1] + bottom)
    segGlabel = (segGlabel[0] + left, segGlabel[1] + bottom)
    segDlabel = (segDlabel[0] + left, segDlabel[1] + bottom)
    segBlabel = (segBlabel[0] + left, segBlabel[1] + bottom)
    segClabel = (segClabel[0] + left, segClabel[1] + bottom)
    segElabel = (segElabel[0] + left, segElabel[1] + bottom)
    segFlabel = (segFlabel[0] + left, segFlabel[1] + bottom)

    # Fill based on digit parameter
    fillA = segcolor if str(digit).lower() in ['2', '3', '5', '6', '7', '8', '9', '0', 'a', 'c', 'e', 'f'] else None
    fillB = segcolor if str(digit).lower() in ['1', '2', '3', '4', '7', '8', '9', '0', 'a', 'd'] else None
    fillC = segcolor if str(digit).lower() in ['1', '3', '4', '5', '6', '7', '8', '9', '0', 'a', 'b', 'd'] else None
    fillD = segcolor if str(digit).lower() in ['2', '3', '5', '6', '8', '9', '0', 'b', 'c', 'd', 'e'] else None
    fillE = segcolor if str(digit).lower() in ['2', '6', '8', '0', 'a', 'b', 'c', 'd', 'e', 'f'] else None
    fillF = segcolor if str(digit).lower() in ['4', '5', '6', '8', '9', '0', 'a', 'b', 'c', 'e', 'f'] else None
    fillG = segcolor if str(digit).lower() in ['2', '3', '4', '5', '6', '8', '9', 'a', 'b', 'd', 'e', 'f'] else None

    segments: list[SegmentType] = []
    segments.append(SegmentPoly(list(zip(segAx, segAy)), color='gray', fill=fillA, lw=.5))
    segments.append(SegmentPoly(list(zip(segBx, segBy)), color='gray', fill=fillB, lw=.5))
    segments.append(SegmentPoly(list(zip(segCx, segCy)), color='gray', fill=fillC, lw=.5))
    segments.append(SegmentPoly(list(zip(segDx, segDy)), color='gray', fill=fillD, lw=.5))
    segments.append(SegmentPoly(list(zip(segEx, segEy)), color='gray', fill=fillE, lw=.5))
    segments.append(SegmentPoly(list(zip(segFx, segFy)), color='gray', fill=fillF, lw=.5))
    segments.append(SegmentPoly(list(zip(segGx, segGy)), color='gray', fill=fillG, lw=.5))
    if labelsegments:
        segments.append(SegmentText(segAlabel, 'a', align=('center', 'bottom'), fontsize=10))
        segments.append(SegmentText(segGlabel, 'g', align=('center', 'bottom'), fontsize=10))
        segments.append(SegmentText(segDlabel, 'd', align=('center', 'bottom'), fontsize=10))
        segments.append(SegmentText(segBlabel, 'b', align=('left', 'center'), fontsize=10))
        segments.append(SegmentText(segClabel, 'c', align=('left', 'center'), fontsize=10))
        segments.append(SegmentText(segElabel, 'e', align=('right', 'center'), fontsize=10))
        segments.append(SegmentText(segFlabel, 'f', align=('right', 'center'), fontsize=10))

    if decimal:
        dotrad = .15
        segments.append(SegmentCircle((left+seglen+segw+dotrad/2, bottom),
                                      radius=dotrad, color='gray', fill=segcolor, lw=.5))
    return segments


class SevenSegment(Ic):
    ''' A seven-segment display digit.

        Keyword Args:
            decimal: Show decimal point segment [default: False]
            digit: Number to display [default: 8]
            segcolor: Color of segments [default: red]
            tilt: Tilt angle in degrees [default: 10]
            labelsegments: Add a-g labels to each segment [default: True]
            anode: Add common anode pin [default: False]
            cathode: Add common cathode pin [default: False]

        Anchors:
            * a
            * b
            * c
            * d
            * e
            * f
            * g
            * dp
            * cathode
            * anode
    '''
    _element_defaults = {
        'decimal': False,
        'digit': 8,
        'segcolor': 'red',
        'tilt': 10,
        'labelsegments': True,
        'anode': False,
        'cathode': False,
        'pinspacing': 0.6
    }
    def __init__(self, *,
                 decimal: Optional[bool] = None,
                 digit: Optional[int | str] = None,
                 segcolor: Optional[str] = None,
                 tilt: Optional[float] = None,
                 labelsegments: Optional[bool] = None,
                 anode: Optional[bool] = None,
                 cathode: Optional[bool] = None,
                 **kwargs) -> None:

        dec = self.params['decimal']
        boxwidth = 3
        if dec:
            slots = '8'
            boxheight = 5.9
        else:
            slots = '7'
            boxheight = 5.3
        size = self.params.get('size', (boxwidth, boxheight))

        pins = [IcPin(pin='a', side='left', slot=f'{7+dec}/{slots}', anchorname='a'),
                IcPin(pin='b', side='left', slot=f'{6+dec}/{slots}', anchorname='b'),
                IcPin(pin='c', side='left', slot=f'{5+dec}/{slots}', anchorname='c'),
                IcPin(pin='d', side='left', slot=f'{4+dec}/{slots}', anchorname='d'),
                IcPin(pin='e', side='left', slot=f'{3+dec}/{slots}', anchorname='e'),
                IcPin(pin='f', side='left', slot=f'{2+dec}/{slots}', anchorname='f'),
                IcPin(pin='g', side='left', slot=f'{1+dec}/{slots}', anchorname='g')]
        if dec:
            pins.append(IcPin(pin='dp', side='left', slot=f'1/{slots}', anchorname='dp'))
        if self.params['anode']:
            pins.append(IcPin(pin='ca', side='top', anchorname='anode'))
        if self.params['cathode']:
            pins.append(IcPin(pin='cc', side='bottom', anchorname='cathode'))

        super().__init__(pins=pins, size=size)

        left = 0.8
        seglen = 1.5
        bot = (boxheight-seglen*2)/2

        segments = sevensegdigit(left=left, bottom=bot,
                                 decimal=dec,
                                 digit=self.params['digit'],
                                 segcolor=self.params['segcolor'],
                                 tilt=self.params['tilt'],
                                 labelsegments=self.params['labelsegments'])
        self.segments.extend(segments)



from .flow import Box, RoundBox, Subroutine, Data, Start, Ellipse, Decision, Connect, Process, RoundProcess
from .flow import Terminal, Circle, State, StateEnd
from ..elements import Arrow, Arrowhead, Line, Dot, Wire, Arc2, Arc3, ArcZ, ArcN, ArcLoop

__all__ = ['Box', 'RoundBox', 'Subroutine', 'Data', 'Start', 'Ellipse', 'Decision', 'Connect',
           'Process', 'RoundProcess', 'Terminal', 'Circle', 'State', 'StateEnd', 'Arrow',
           'Arrowhead', 'Line', 'Dot', 'Wire', 'Arc2', 'Arc3', 'ArcZ', 'ArcN', 'ArcLoop']
